#!/usr/bin/env python3
"""Unit test for trepan.lib.bytecode"""
import inspect

from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE

from trepan.lib.bytecode import is_def_stmt, op_at_frame, stmt_contains_opcode


def test_contains_make_function():
    frame = inspect.currentframe()
    co = frame.f_code
    lineno = frame.f_lineno
    assert not stmt_contains_opcode(co, lineno, "MAKE_FUNCTION")
    return


def test_op_at_frame():
    frame = inspect.currentframe()
    if IS_PYPY or PYTHON_VERSION_TRIPLE >= (3, 9):
        if PYTHON_VERSION_TRIPLE >= (3, 11):
            call_opcode = "CACHE"
        elif PYTHON_VERSION_TRIPLE > (3, 9):
            call_opcode = "CALL_FUNCTION"
        else:
            call_opcode = "CALL_METHOD"
    else:
        call_opcode = "CALL_FUNCTION"

    assert call_opcode == op_at_frame(frame)
    return


def test_is_def_frame():
    # Not a "def" statement because frame is wrong spot
    frame = inspect.currentframe()
    assert not is_def_stmt("foo(): pass", frame)
    return



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013, 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.lib.pp import pp
from trepan.processor.complete import complete_identifier


class PrettyPrintCommand(DebuggerCommand):
    """**pp** *expression*

    Pretty-print the value of the expression.

    Simple arrays are shown columnized horizontally. Other values are printed
    via *pprint.pformat()*.

    See also:
    ---------

    `pr` and `examine` for commands which do more in the way of
    formatting."""

    short_help = "Pretty print value of expression EXP"

    complete = complete_identifier

    DebuggerCommand.setup(locals(), category="data", min_args=1)

    def run(self, args):
        arg = " ".join(args[1:])
        val = self.proc.eval(arg)
        pp(val, self.settings["width"], self.msg_nocr, self.msg)
        return False

    pass


if __name__ == "__main__":
    import inspect
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    cp.curframe = inspect.currentframe()
    command = PrettyPrintCommand(cp)
    me = list(range(10))
    command.run(["pp", "me"])
    me = list(range(100))
    command.run(["pp", "me"])
    import sys  # noqa

    command.run(["pp", "sys.modules.keys()"])
    me = "fooled you"
    command.run(["pp", "locals()"])
    pass



"""
Functional test of debugger "break" command.
"""
import unittest
from test.functional.fn_helper import compare_output, strarray_setup


class TestBreak(unittest.TestCase):
    def test_break_on_function(self):
        print("test ", __file__, "break_on_function skipped")
        return

        ##############################
        # We had a bug where 'next' (no number) after
        # a hit breakpoint wouldn't 'next'. So test for that
        # in addition to the break on a function name.
        cmds = ["break foo", "continue", "next", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################

        def foo():
            print("foo here")
            return

        foo()
        y = 6  # NOQA
        ##############################
        d.core.stop()

        out = ["xx def foo():\n(test-break.py:18): foo", "-- print 'foo here'\n"]
        compare_output(self, out, d)

        # Try a break with a module name
        import os

        cmds = ["break os.path.join", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        p = os.path.join("a", "b")  # NOQA
        ##############################
        d.core.stop()
        out = ['-- p = os.path.join("a", "b")', "xx def join(a, *p):"]
        compare_output(self, out, d)
        return

    def test_break_at_line_number(self):
        print("test", __file__, "break_at_line_number skipped")
        return
        import inspect

        curframe = inspect.currentframe()
        cmds = ["break %d" % (curframe.f_lineno + 7), "continue", "c"]  # 1
        d = strarray_setup(cmds)  # 2
        d.core.start()  # 3
        ##############################        # 4...
        x = 5  # NOQA
        y = 6  # NOQA
        z = 7  # NOQA
        ##############################
        d.core.stop()
        out = [
            "-- x = 5\nBreakpoint 1 set at line 53 of file test-break.py",
            "xx z = 7\n",
        ]
        compare_output(self, out, d)
        return

    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2015 Rocky Bernstein
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#    02110-1301 USA.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowHighlight(Mbase_subcmd.DebuggerSubcommand):
    """**show highlight**

    Show whether we use terminal highlighting.

    See also:
    --------

    `set highlight`"""

    short_help = "Show if we use terminal highlight"

    def run(self, args):
        val = self.settings["highlight"]
        if "plain" == val:
            mess = "output set to not use terminal escape sequences"
        elif "light" == val:
            mess = (
                "output set for terminal with escape sequences "
                "for a light background"
            )
        elif "dark" == val:
            mess = (
                "output set for terminal with escape sequences " "for a dark background"
            )
        else:
            self.errmsg("Internal error: incorrect highlight setting %s" % val)
            return
        self.msg(mess)
        return

    pass



# -*- coding: iso-8859-1 -*-
#   Copyright (C) 2013-2015, 2023-2024 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import codecs
import os
import os.path as osp
import sys
from optparse import OptionParser

from pygments.styles import STYLE_MAP

import trepan.api
from trepan.clifns import path_expanduser_abs
from trepan.inout.output import DebuggerUserOutput
from trepan.lib.file import readable


def default_configfile(base_filename: str) -> str:
    """Return fully expanded configuration filename location for
    base_filename. python2 and  python3 debuggers share the same
    directory: ~/.config/trepan.py
    """
    file_dir = osp.join(os.environ.get("HOME", "~"), ".config", "trepanpy")
    file_dir = path_expanduser_abs(file_dir)

    if not osp.isdir(file_dir):
        os.makedirs(file_dir, mode=0o755)
    return osp.join(file_dir, base_filename)


def add_startup_file(dbg_initfiles: list):
    """Read debugger startup file(s): both python code and
    debugger profile to dbg_initfiles."""

    startup_python_file = default_configfile("profile.py")

    if readable(startup_python_file):
        with codecs.open(startup_python_file, "r", encoding="utf8") as fp:
            exec(fp.read())

    startup_trepan_file = default_configfile("profile")
    if readable(startup_trepan_file):
        if startup_trepan_file not in dbg_initfiles:
            dbg_initfiles.append(startup_trepan_file)
        pass
    return


def process_options(pkg_version: str, sys_argv: str, option_list=None):
    """Handle debugger options. Set ``option_list`` if you are writing
    another main program and want to extend the existing set of debugger
    options.

      Note that ``sys.argv`` must contain at least one element, the program name.

    The options dictionary from optparser is returned. sys_argv is
    also updated."""

    usage_str = """%prog [debugger-options] [python-script [script-options...]]

    Runs the extended python debugger"""

    assert (
        len(sys_argv) > 0
    ), "sys_argv parameter must contain at least one string, the program name."
    # serverChoices = ('TCP','FIFO', None)

    optparser = OptionParser(
        usage=usage_str,
        option_list=option_list,
        version=f"%prog version {pkg_version}",
    )

    optparser.add_option(
        "-X",
        "--trace",
        dest="linetrace",
        action="store_true",
        default=False,
        help="Show lines before executing them. This option also sets --batch",
    )
    optparser.add_option(
        "-F",
        "--fntrace",
        dest="fntrace",
        action="store_true",
        default=False,
        help="Show functions before executing them. This option also sets --batch",
    )
    optparser.add_option(
        "--basename",
        dest="basename",
        action="store_true",
        default=False,
        help="Filenames strip off basename, (e.g. for regression tests)",
    )
    #     optparser.add_option("--batch", dest="noninteractive",
    #                          action="store_true", default=False,
    #                          help="Don't run interactive commands shell on "+
    #                          "stops.")
    optparser.add_option(
        "--client",
        dest="client",
        action="store_true",
        help="Connect to an existing debugger process "
        "started with the --server option. "
        "See options for client.",
    )
    optparser.add_option(
        "-x",
        "--command",
        dest="command",
        action="store",
        type="string",
        metavar="FILE",
        help="Execute commands from FILE.",
    )
    optparser.add_option(
        "--cd",
        dest="cd",
        action="store",
        type="string",
        metavar="DIR",
        help="Change current directory to DIR.",
    )
    optparser.add_option(
        "--confirm",
        dest="confirm",
        action="store_true",
        default=True,
        help="Confirm potentially dangerous operations",
    )
    optparser.add_option(
        "--dbg_trepan",
        dest="dbg_trepan",
        action="store_true",
        default=False,
        help="Debug the debugger",
    )
    optparser.add_option(
        "--different",
        dest="different",
        action="store_true",
        default=True,
        help="Consecutive stops should have different positions",
    )
    #     optparser.add_option("--error", dest="errors", metavar='FILE',
    #                          action="store", type='string',
    #                          help="Write debugger's error output "
    #                          + "(stderr) to FILE")
    optparser.add_option(
        "-e",
        "--exec",
        dest="execute",
        type="string",
        help="list of debugger commands to " + "execute. Separate the commands with ;;",
    )

    optparser.add_option(
        "-H",
        "--host",
        dest="host",
        default="127.0.0.1",
        action="store",
        type="string",
        metavar="IP-OR-HOST",
        help="connect IP or host name. Only valid if --client option given.",
    )

    optparser.add_option(
        "--highlight",
        dest="highlight",
        action="store",
        type="string",
        metavar="{light|dark|plain}",
        default="light",
        help="Use syntax and terminal highlight output. 'plain' is no highlight",
    )

    optparser.add_option(
        "--private",
        dest="private",
        action="store_true",
        default=False,
        help="Don't register this as a global debugger",
    )

    optparser.add_option(
        "--main",
        dest="main",
        action="store_true",
        default=True,
        help="First stop should be in __main__",
    )
    optparser.add_option(
        "--no-main",
        dest="main",
        action="store_false",
        default=True,
        help="First stop should be in __main__",
    )
    optparser.add_option(
        "--post-mortem",
        dest="post_mortem",
        action="store_true",
        default=True,
        help="Enter debugger on an uncaught (fatal) exception",
    )

    optparser.add_option(
        "--no-post-mortem",
        dest="post_mortem",
        action="store_false",
        default=True,
        help="Don't enter debugger on an uncaught (fatal) exception",
    )

    optparser.add_option(
        "-n",
        "--nx",
        dest="noexecute",
        action="store_true",
        default=False,
        help="Don't execute commands found in any initialization files",
    )

    optparser.add_option(
        "-o",
        "--output",
        dest="output",
        metavar="FILE",
        action="store",
        type="string",
        help="Write debugger's output (stdout) to FILE",
    )
    optparser.add_option(
        "-P",
        "--port",
        dest="port",
        default=1027,
        action="store",
        type="int",
        help="Use TCP port number NUMBER for out-of-process connections.",
    )

    optparser.add_option(
        "--server",
        dest="server",
        action="store_true",
        help="Out-of-process server connection mode",
    )

    optparser.add_option(
        "--style",
        dest="style",
        action="store",
        type="string",
        metavar="*pygments-style*",
        default=None,
        help="Pygments style; 'none' uses 8-color rather than 256-color terminal",
    )

    optparser.add_option(
        "--sigcheck",
        dest="sigcheck",
        action="store_true",
        default=False,
        help="Set to watch for signal handler changes",
    )
    optparser.add_option(
        "-t",
        "--target",
        dest="target",
        help=(
            "Specify a target to connect to. Arguments"
            " should be of form, 'protocol address'."
        ),
    ),
    optparser.add_option(
        "--from_ipython",
        dest="from_ipython",
        action="store_true",
        default=False,
        help="Called from inside ipython",
    )

    # annotate option produces annotations, used in trepan.el for a
    # better emacs integration. Annotations are similar in purpose to
    # those of GDB (see that manual for a description), although the
    # syntax is different.  they have the following format:
    #
    # ^Z^Zannotation-name
    # <arbitrary text>
    # ^Z^Z
    #
    # where ^Z is the ctrl-Z character, and "annotation" is the name of the
    # annotation. A line with only two ^Z ends the annotation (no nesting
    # allowed). See trepan.el for the usage
    optparser.add_option(
        "--annotate", default=0, type="int", help="Use annotations to work inside emacs"
    )

    # Set up to stop on the first non-option because that's the name
    # of the script to be debugged on arguments following that are
    # that scripts options that should be left untouched.  We would
    # not want to interpret an option for the script, e.g. --help, as
    # one of our own, e.g. --help.

    optparser.disable_interspersed_args()

    sys.argv = list(sys_argv)
    (opts, sys.argv) = optparser.parse_args(sys_argv[1:])
    dbg_opts = {"from_ipython": opts.from_ipython}

    # Handle debugger startup command files: --nx (-n) and --command.
    dbg_initfiles = []
    if not opts.noexecute:
        add_startup_file(dbg_initfiles)

    # As per gdb, first we execute user initialization files, and then
    # we execute any file specified via --command.
    if opts.command:
        dbg_initfiles.append(opts.command)
        pass

    dbg_opts["proc_opts"] = {"initfile_list": dbg_initfiles}

    if opts.cd:
        os.chdir(opts.cd)
        pass

    if opts.output:
        try:
            dbg_opts["output"] = DebuggerUserOutput(opts.output)
        except IOError:
            _, xxx_todo_changeme, _ = sys.exc_info()
            (errno, strerror) = xxx_todo_changeme.args
            print(f"I/O in opening debugger output file {opts.output}")
            print(f"error({errno}): {strerror}")
        except Exception:
            print(f"Unexpected error in opening debugger output file {opts.output}")
            print(sys.exc_info()[0])
            sys.exit(2)
            pass
        pass

    return opts, dbg_opts, sys.argv


def postprocess_options(dbg, opts):
    """Handle options (``opts`) that feed into the debugger (``dbg``)"""
    # Set dbg.settings['printset']
    print_events = []
    if opts.fntrace:
        print_events = ["c_call", "c_return", "call", "return"]
    if opts.linetrace:
        print_events += ["line"]
    if len(print_events):
        dbg.settings["printset"] = frozenset(print_events)
        pass

    for setting in ("annotate", "basename", "different"):
        dbg.settings[setting] = getattr(opts, setting)
        pass

    if getattr(opts, "highlight"):
        dbg.settings["highlight"] = opts.highlight
    else:
        dbg.settings["highlight"] = "plain"

    if getattr(opts, "style") and opts.style != "none":
        style_names = sorted(list(STYLE_MAP.keys()))
        if opts.style in style_names:
            dbg.settings["style"] = opts.style
        else:
            sys.stderr.write(
                f"""Pygments style "{opts.style}" listed. --style option ignored.\n"""
                f"""Use 'set style" to see valid style and change style.\n"""
            )

    # Normally we want to set trepan.api.debugger_obj so that one can
    # put trepan.debugger breakpoints in a program and not have more
    # than one debugger running. More than one debugger may confuse
    # users, e.g. set different might stop at the same line once for
    # each debugger.
    if not opts.private:
        trepan.api.debugger_obj = dbg
        pass

    #     if opts.errors:
    #         try:
    #             dbg.stderr = open(opts.errors, 'w')
    #         except IOError, (errno, strerror):
    #             print "I/O in opening debugger output file %s" % opts.errors
    #             print "error(%s): %s" % (errno, strerror)
    #         except ValueError:
    #             print "Could not convert data to an integer."
    #         except Exception:
    #             print "Unexpected error in opening debugger output "
    #                   "file %s" % opts.errors
    #             print sys.exc_info()[0]
    #             sys.exit(2)

    #     if opts.execute:
    #         dbg.cmdqueue = list(opts.execute.split(';;'))

    if opts.post_mortem:
        from trepan.api import debugger_on_post_mortem

        debugger_on_post_mortem()
        pass
    return


# Demo it
if __name__ == "__main__":
    import pprint

    def doit(version, arg_str):
        print(f"options '{arg_str}'")
        args = arg_str.split()
        opts, dbg_opts, sys_argv = process_options(version, args)
        pp.pprint(vars(opts))
        print("")
        return

    pp = pprint.PrettyPrinter(indent=4)
    doit("1.1", "__file__")
    doit("1.2", f"{__file__} foo bar")
    doit("1.3", f"{__file__} --server")
    doit("1.3", f"{__file__} --command {__file__} bar baz")
    doit("1.4", f"{__file__} --server --client")
    doit("1.5", f"{__file__} --style=emacs")
    doit("1.6", f"{__file__} --help")  # exits, so must be last
    pass



import unittest
from test.functional.fn_helper import compare_output, strarray_setup


class TestSkip(unittest.TestCase):
    def test_skip(self):
        # See that we can skip without parameter. (Same as 'skip 1'.)
        cmds = ["skip", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 4
        x = 5
        y = 7  # NOQA
        ##############################
        d.core.stop()
        out = ["-- x = 4", "-- x = 5"]  # x = 4 is shown in prompt, but not *run*.
        compare_output(self, out, d)
        self.assertEqual(5, x)  # Make sure lines were skipped.

        # See that we can skip with a count value
        cmds = ["skip 2", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 10
        x = 9
        z = 7  # NOQA
        ##############################
        d.core.stop(options={"remove": True})
        out = [
            "-- x = 10",
            "-- z = 7  # NOQA",
        ]  # x = 10 is shown in prompt, but not run.
        compare_output(self, out, d)
        self.assertEqual(5, x)  # Make sure x = 10, 9 were skipped.
        return



"""Unit test for trepan.processor.subcmd"""
import unittest

from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.command.mock import MockDebugger
from trepan.processor.subcmd import Subcmd


class MyCommand(DebuggerCommand):
    """Trepan processor Debugger Command base"""

    category = "data"
    min_args = 0
    max_args = 5
    name_aliases = "mycommand"

    def __init__(self):
        self.name = "test"
        return

    def run(self, args):
        print("test command run")

    pass


class MySubcommand:
    """Doc string for test testing subcommand"""

    def __init__(self):
        self.name = "testing"
        return

    short_help = "This is short help for test testing"
    min_abbrev = 4
    in_list = True

    def run(self, args):
        print("test testing run")

    pass


class TestSubcommand(unittest.TestCase):
    def setUp(self):
        self.d = MockDebugger()
        self.mycmd = MyCommand()
        self.mycmd.debugger = self.d
        self.mycmd.proc = self.d.core.processor
        self.mycmdMgr = Subcmd("me", self.mycmd)
        self.testsub = MySubcommand()
        self.mycmdMgr.add(self.testsub)

    def test_lookup(self):
        """Test processor.subcmd.lookup()"""
        for prefix, expected in (
            ("tes", "None"),  # Too few chars
            ("test", "testing"),  # A valid abbrev
            ("testing", "testing"),  # equal name
            ("testing1", "None"),
        ):  # Too long
            x = self.mycmdMgr.lookup(prefix)
            if x:
                s = x.name
            else:
                s = "None"
            self.assertEqual(
                expected,
                s,
                ("prefix %s, expected: %s result: %s" % (prefix, expected, s)),
            )
            pass
        return

    def test_list(self):
        """Test processor.subcmd.list()"""
        self.assertEqual(["testing"], self.mycmdMgr.list())
        testsub2 = MySubcommand()
        testsub2.name = "foobar"
        self.mycmdMgr.add(testsub2)
        self.assertEqual(["foobar", "testing"], self.mycmdMgr.list())

    pass


if __name__ == "__main__":
    unittest.main()



"""Unit test for trepan.options"""

from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.options import postprocess_options, process_options


def test_options():
    """
    Test trepan.options.process_options() and trepan.options.postprocess.
    """

    # Setup.
    option_key_set = {
        "annotate",
        "basename",
        "cd",
        "client",
        "command",
        "confirm",
        "dbg_trepan",
        "different",
        "execute",
        "fntrace",
        "from_ipython",
        "highlight",
        "host",
        "linetrace",
        "main",
        "noexecute",
        "output",
        "port",
        "post_mortem",
        "private",
        "server",
        "sigcheck",
        "style",
        "target",
    }
    dbg, _ = setup_unit_test_debugger()
    dbg_opts_set = {"proc_opts", "from_ipython"}

    # Test with no options. See that we have the some expected
    # keys getting set.
    opts, dbg_opts, sys_argv = process_options("1.0", [__file__])
    diff_set = option_key_set - set(vars(opts).keys())
    assert diff_set == set(), "expecting at least these options keys set"
    assert (
        dbg_opts_set - set(dbg_opts.keys())
    ) == set(), "expecting at least these processor keys set"

    # Try with more than one option, a boolean option and a string option.
    arg_str = f"{__file__} --fntrace --cd=/tmp"
    opts, dbg_opts, sys_argv = process_options("1.1", arg_str.split())
    assert opts.cd == "/tmp"
    postprocess_options(dbg, opts)
    assert dbg.settings["printset"] == frozenset(
        ["c_call", "c_return", "call", "return"]
    )

    # Try with an invalid style option and see that it is
    # rejected in postprocess option
    arg_str = f"{__file__} --style=fafdsaXYZZY"
    opts, dbg_opts, sys_argv = process_options("1.3", arg_str.split())
    postprocess_options(dbg, opts)
    # FIXME figure out a reasonable test for postprocess_options()



# -*- coding: utf-8 -*-
"""Deparsing Routines"""

import sys
import tempfile
from hashlib import sha1
from io import StringIO
from typing import Callable

import pyficache
from xdis import PYTHON_VERSION_TRIPLE

if PYTHON_VERSION_TRIPLE >= (3, 9):
    pass
else:
    if (3, 7) <= PYTHON_VERSION_TRIPLE:
        try:
            from decompyle3.semantics.fragments import code_deparse, deparsed_find
            from decompyle3.semantics.linemap import code_deparse_with_map
        except ImportError:
            from uncompyle6.semantics.fragments import code_deparse, deparsed_find
            from uncompyle6.semantics.linemap import code_deparse_with_map
    else:
        from uncompyle6.semantics.fragments import code_deparse, deparsed_find
        from uncompyle6.semantics.linemap import code_deparse_with_map

    # FIXME remap filename to a short name.

    deparse_cache = {}

    def deparse_and_cache(co, errmsg_fn, tempdir=None):
        # co = proc_obj.curframe.f_code
        out = StringIO()
        deparsed = deparse_cache.get(co, None)
        if not deparsed or not hasattr(deparsed, "source_linemap"):
            try:
                deparsed = code_deparse_with_map(co, out)
            except Exception:
                errmsg_fn(str(sys.exc_info()[0]))
                errmsg_fn(f"error in deparsing code: {co.co_filename}")
                return None, None

            deparse_cache[co] = deparsed

        text = out.getvalue()
        linemap = [
            (line_no, deparsed.source_linemap[line_no])
            for line_no in sorted(deparsed.source_linemap.keys())
        ]

        # FIXME: DRY code with version in cmdproc.py print_location

        name_for_code = sha1(co.co_code).hexdigest()[:6]
        prefix = "deparsed-"
        fd = tempfile.NamedTemporaryFile(
            suffix=".py", prefix=prefix, dir=tempdir, delete=False
        )
        with fd:
            fd.write(text.encode("utf-8"))
            map_line = f"\n\n# {linemap}"
            fd.write(map_line.encode("utf-8"))
            remapped_file = fd.name
        fd.close()
        # FIXME remap filename to a short name.
        pyficache.remap_file_lines(name_for_code, remapped_file, linemap)
        return remapped_file, name_for_code

    def deparse_offset(co, name: str, last_i: int, errmsg_fn: Callable) -> tuple:
        node_info = None
        deparsed = deparse_cache.get(co, None)
        if not deparsed or not hasattr(deparsed, "offsets"):
            out = StringIO()
            try:
                # FIXME: cache co
                deparsed = code_deparse(co, out)
            except Exception:
                print(sys.exc_info()[1])
                if errmsg_fn:
                    errmsg_fn(sys.exc_info()[1])
                    errmsg_fn("error in deparsing code")
            deparse_cache[co] = deparsed
        try:
            node_info = deparsed_find((name, last_i), deparsed, co)
        except Exception:
            if errmsg_fn:
                errmsg_fn(sys.exc_info()[1])
                errmsg_fn("error in deparsing code at offset %d" % last_i)

        if not node_info:
            node_info = deparsed_find((name, last_i), deparsed, co)
        return deparsed, node_info


# Demo it
if __name__ == "__main__":
    import inspect

    def msg(msg_str):
        print(msg_str)
        return

    def errmsg(msg_str):
        msg("*** " + msg_str)
        return

    curframe = inspect.currentframe()
    mapped_name, name_for_code = deparse_and_cache(curframe.f_code, errmsg)
    print(pyficache.getline(mapped_name, 7))
    # mapped_name, name_for_code = deparse_offset(curframe.f_code,
    #                                             curframe.f_code.co_name,
    #                                             curframe.f_lasti, errmsg)
    # print(pyficache.getline(mapped_name, 7))



"""
Unit test of trepan.api
"""

import sys

import pytest
import tracer
from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE

from trepan.api import run_call, run_eval
from trepan.inout.stringarray import StringArrayInput, StringArrayOutput
from trepan.lib.default import DEBUGGER_SETTINGS


def plus5(n: int) -> int:
    return n + 5


@pytest.mark.skipif(
    IS_PYPY and PYTHON_VERSION_TRIPLE < (3, 7, 0),
    reason="PyPy strings work differently",
)
def test_run_xxx():
    """
    test run_eval() and run_call()
    """
    settings = dict(DEBUGGER_SETTINGS)
    settings.update({"trace": True, "printset": tracer.ALL_EVENTS})
    debugger_input = StringArrayInput()
    debugger_output = StringArrayOutput()
    debug_opts = {
        "step_ignore": -1,
        "settings": settings,
        "input": debugger_input,
        "output": debugger_output,
    }
    print('Issuing: run_eval("1+2")')
    print(run_eval("1+2", debug_opts=debug_opts))
    print(debugger_output.output)
    start_lineno = "1" if sys.version_info[:2] < (3, 11) else "0"

    assert debugger_output.output[0:3] == [
        "call - <string>:%s" % start_lineno,
        "line - <string>:1",
        "return - <string>:1, 3 ",
    ]

    debugger_input = StringArrayInput([""])
    debugger_output = StringArrayOutput()
    debug_opts = {
        "step_ignore": -1,
        "settings": settings,
        "input": debugger_input,
        "output": debugger_output,
    }
    print('Issuing: run_call(foo, 3")')
    x = run_call(plus5, 3, debug_opts=debug_opts)
    assert x == 8
    call_lineno = "16"
    for i, (prefix, suffix) in enumerate(
        (
            ("call - ", "test/unit/test_api.py:%s" % call_lineno),
            ("line - ", "test/unit/test_api.py:17"),
            ("return - ", "test/unit/test_api.py:17, 8 "),
        )
    ):
        assert debugger_output.output[i].startswith(prefix), (
            prefix,
            suffix,
            debugger_output.output[i],
        )
        assert debugger_output.output[i].endswith(suffix), (
            prefix,
            suffix,
            debugger_output.output[i],
        )



"""Unit test for trepan.lib.core"""
import os
import sys

from trepan.lib.core import TrepanCore


class MockProcessor:
    pass


def test_is_next_stop():
    # We assume there's at least one command
    opts = {"processor": MockProcessor()}
    dc = TrepanCore(None, opts=opts)
    dc.step_ignore = 1
    assert not dc._is_step_next_stop("line")
    assert 0 == dc.step_ignore
    assert dc._is_step_next_stop("line")
    assert 0 == dc.step_ignore

    return


def test_canonic():
    opts = {"processor": MockProcessor()}
    dc = TrepanCore(None, opts=opts)
    s = "<string>"
    assert s == dc.canonic(s), "canonic should not have changed string"
    if sys.platform != "win32":
        assert (
            os.path.sep == dc.canonic(__file__)[0]
        ), "canonic should produce an absolute file"
    return



# -*- coding: utf-8 -*-
""" Mock classes used for testing which in turn really come from
trepan.processor.command.mock """

import inspect
from io import StringIO
from typing import List

from trepan import debugger
from trepan.inout.input import DebuggerUserInput
from trepan.interfaces.user import UserInterface
from trepan.processor.command import mock

dbg_setup = mock.dbg_setup

errmsgs: List[str] = []


def errmsg(msg_str: str):
    """
    error message collection routine for unit testing.
    """
    global errmsgs
    errmsgs.append(msg_str)
    return


msgs: List[str] = []


def msg(msg_str: str):
    """
    message collection routine for unit testing.
    """
    global msgs
    msgs.append(msg_str)
    return


def reset_output():
    """
    Reset error messages and messages in advance of testing.
    """
    global errmsgs
    errmsgs = []
    global msgs
    msgs = []


def setup_unit_test_debugger(
    input_string: str = "", debugger_name: str = "trepan3k-unit-pytest"
) -> tuple:
    """
    Creates a trepan debugger object with  input bound to the
    StringIO text from "input_string", and no history tracking,
    The debugger and it core processor are returned
    """

    inp = StringIO(input_string)
    user_input = DebuggerUserInput(inp)

    def no_history():
        return False

    user_input.use_history = no_history
    interface_opts = UserInterface(
        inp=user_input, opts={"debugger_name": debugger_name}
    )

    d = debugger.Trepan(opts={"interface": interface_opts})
    cp = d.core.processor
    cp.curframe = inspect.currentframe().f_back
    cp.list_filename = cp.curframe.f_code.co_filename
    return d, cp


def readline_yes(prompt=None) -> str:
    return "Y"


def readline_no(prompt=None) -> str:
    return "N"



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2010, 2013-2015, 2020-2021, 2024
#   Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from glob import glob
from typing import Any


def option_set(options: dict, value, default_options) -> Any:
    """
    If ``value`` is found in ``options``, return that, otherwise return the value
    from ``defaulit_options``.
    """
    # Yes, there is probably some fancy dictionary merge operation, that will do this,
    # as a one-shot but this code is simple and clear.
    if not options or value not in options:
        return default_options.get(value)
    else:
        return options.get(value)
    return None  # Not reached


def bool2YN(b: bool) -> str:
    """
    Turn a bool into the string "Y" or "N".
    """
    return "Y" if b else "N"


def wrapped_lines(msg_part1: str, msg_part2: str, width: int) -> str:
    """
    if ``msg_part1`` concatenated with ``msg_part2`` is larger than
    ``width`` then concatenate the strings with a new line and tab inserted
    between the strings. Otherwise, just concatenate with a space between the
    two strings.
    """
    if len(msg_part1) + len(msg_part2) + 1 > width:
        return msg_part1 + "\n\t" + msg_part2
    else:
        return msg_part1 + " " + msg_part2


def pretty_modfunc_name(s) -> str:
    if s == "<module>":
        # FIXME:
        # Pick replace with something more custom to modname?
        return str(s)
    else:
        return str(s) + "()"


def pyfiles(callername, level=2):
    "All python files caller's dir without the path and trailing .py"
    d = os.path.dirname(callername)
    # Get the name of our directory.
    # A glob pattern that will get all *.py files but not __init__.py
    glob(os.path.join(d, "[a-zA-Z]*.py"))
    py_files = glob(os.path.join(d, "[a-zA-Z]*.py"))
    return [os.path.basename(filename[0:-3]) for filename in py_files]


# Demo it
if __name__ == "__main__":
    TEST_OPTS = {"a": True, "b": 5, "c": None}

    def get_option(key):
        return option_set(opts, key, TEST_OPTS)

    opts = {"d": 6, "a": False}
    for opt in ["a", "b", "c", "d"]:
        print(opt, get_option(opt))
        pass
    for b in [True, False]:
        print(bool2YN(b))
    pass

    print(wrapped_lines("hi", "there", 80))
    print(wrapped_lines("hi", "there", 5))
    print(pyfiles(__file__))
    pass



"""Unit test for trepan.lib.sighandler"""
import signal

from trepan.lib import sighandler as Msig


def test_YN():
    for expect, b in (("Yes", True), ("No", False)):
        expect == Msig.yes_or_no(b)
        pass
    return


def test_canonic_signame():
    for expect, name_num in (
        ("SIGTERM", "15"),
        ("SIGTERM", "-15"),
        ("SIGTERM", "term"),
        ("SIGTERM", "sigterm"),
        ("SIGTERM", "TERM"),
        (None, "300"),
        (None, "bogus"),
    ):
        expect == Msig.canonic_signame(name_num), f"name_num: {name_num}"
        pass
    pass


def test_lookup_signame():
    for expect, num in (("SIGTERM", 15), ("SIGTERM", -15), (None, 300)):
        expect == Msig.lookup_signame(num), f"looking up {num}"
        pass
    return


def test_lookup_signum():
    for expect, name in (
        (15, "SIGTERM"),
        (15, "TERM"),
        (15, "term"),
        (None, "nothere"),
    ):
        expect == Msig.lookup_signum(name), f"looking up name {name}"
        pass
    return


def test_lookup_signame_signum():
    ignore = ["ITIMER_VIRTUAL", "ITIMER_REAL", "ITIMER_PROF"]
    for signum in range(signal.NSIG):
        signame = Msig.lookup_signame(signum)
        if signame is not None and signame not in ignore:
            signum == Msig.lookup_signum(signame), f"looking up name {signame}"
            # Try without the SIG prefix
            signum == Msig.lookup_signum(signame[3:])
            pass
        pass
    return



"""Unit test for trepan.processor.command.list"""

import os  # NOQA
from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.command.list import ListCommand


def test_list_command():
    listsize = 8
    errors = []
    msgs = []

    def errmsg(msg_str: str):
        errors.append(msg_str)
        return

    def msg(msg_str: str):
        msgs.append(msg_str)
        return

    def print_lines():
        for line in msgs:
            print(line)
        for line in errors:
            print(line)

    def check_lines(nums: list):
        j = 0
        # print_lines()

        if len(nums) != len(msgs):
            print_lines()
            assert not f"len(msg): {len(msgs)} vs. len(check): {len(nums)}"
            return

        for i in nums:
            assert ("%3d" % i) == msgs[j][0:3]
            j += 1
            pass
        return

    def clear_run(args):
        cmd.proc.current_command = " ".join(args)
        cmd.run(args)

    def clear_run_check(args, nums):
        clear_run(args)
        check_lines(nums)
        return

    def clear_run_checksize(args):
        clear_run(args)
        assert listsize == len(msgs) - 1
        return

    d, cp = setup_unit_test_debugger()
    cmd = ListCommand(cp)
    cmd.msg = msg
    cmd.errmsg = errmsg
    d.settings["listsize"] = listsize
    d.settings["highlight"] = "plain"

    # Simple list command.
    clear_run_check(["list"], list(range(1, listsize + 1)))
    # Check 2nd set of consecutive lines
    msgs = []
    clear_run_check(["list"], list(range(listsize + 1, (2 * listsize) + 1)))

    # Try going backwards.
    msgs = []
    clear_run_check(["list", "-"], list(range(1, 1 + listsize)))

    # And again. Since we hit the beginning it's the same as before
    msgs = []
    clear_run_check(["list", "-"], list(range(1, 1 + listsize)))

    # With the roll in of a parser, we no longer support
    # expression arithmetic
    # clear_run_check(['list', '4+1'], range(4+1, 4+1+listsize))
    # clear_run_check(['list', __file__+':3', ',4'], list(range(3, 5)))

    # List first last
    msgs = []
    clear_run_check(["list", "10", ",", "20"], list(range(10, 21)))

    # List first count
    msgs = []
    clear_run_check(["list", "10", ",", "5"], list(range(10, 16)))

    # Module
    # BUG? without '1' below the default starts with listsize+1
    msgs = []
    clear_run_check(["os.path", "1"], list(range(1, listsize + 2)))

    # Function
    msgs = []
    clear_run_checksize(["list", "test_list_command()"])

    def foo():  # NOQA
        pass

    msgs = []
    clear_run_checksize(["list", "foo()"])

    # BUG - os.path can be listed.
    # clear_run_check(['os.path', '10', ',5'], list(range(10, 16)))
    # Use a file name
    # my_file = os.path.realpath(__file__)

    msgs = []
    clear_run_check(["list", f"{__file__}:3, 2"], list(range(3, 3 + 2 + 1)))

    msgs = []
    clear_run_check(["list", f"{__file__}:3, 2"], list(range(3, 3 + 2 + 1)))

    msgs = []
    # 4 < 20, so "4" is a length
    clear_run_check(["list", f"{__file__}:20,", "4"], list(range(20, 20 + 4 + 1)))

    # 4 > 3, so "4" is a last line number
    msgs = []
    clear_run_check(["list", f"{__file__}:3 ,", "4"], list(range(3, 4 + 1)))

    msgs = []
    # Explicit length
    clear_run_check(["list", f"{__file__}:3 ,", "+4"], list(range(3, 3 + 4 + 1)))
    return



"""Unit test for "trepan.frame" """
import inspect

from trepan.lib.stack import count_frames


def test_count_frames():
    f = inspect.currentframe()
    frame_count = count_frames(f)
    assert count_frames(f) > 2
    assert frame_count - 1 == count_frames(f.f_back)
    assert frame_count - 1 == count_frames(f, 1)
    return


# FIXME fix up is_exec_stmt() ?
# def test_stack_misc():
#     f = inspect.currentframe()
#     # assert "test_stack_misc" == get_call_function_name(f))
#     assert not is_exec_stmt(f)
#     result = False
#     exec("result = is_exec_stmt(inspect.currentframe())")
#     assert result
#     return



#!/usr/bin/env python3
"""Unit test for trepan.exception"""

from trepan.exception import DebuggerRestart


def test_debugger_restart():
    try:
        raise DebuggerRestart(["a", "b"])
    except DebuggerRestart:
        import sys

        assert ["a", "b"] == sys.exc_info()[1].sys_argv
    else:
        assert False, "raise DebuggerRestart didn't raise"
    return



# -*- coding: utf-8 -*-
#   Copyright (C) 2015 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class SetConfirm(Mbase_subcmd.DebuggerSetBoolSubcommand):
    """**set confirm** [ **on* | **off** ]

    Set confirmation of potentially dangerous operations.

    Some operations are a bit disruptive like terminating the program.
    To guard against running this accidentally, by default we ask for
    confirmation. Commands can also be exempted from confirmation by suffixing
    them with an exclamation mark (!).

    See Also:
    ---------

    `show confirm`"""

    in_list = True
    min_abbrev = len("co")
    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(SetConfirm)
    pass



#!/usr/bin/env python
"""Unit test for trepan.processor.command.p"""

from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.command.p import PCommand


def test_p():
    errors = []
    msgs = []

    def errmsg(msg_text: str):
        errors.append(msg_text)
        return

    def msg(msg_text: str):
        msgs.append(msg_text)
        return

    d, cp = setup_unit_test_debugger()
    cmd = PCommand(cp)
    cmd.msg = msg
    cmd.errmsg = errmsg

    me = 10  # NOQA
    cmd.run([cmd.name, "me"])
    assert "10" == msgs[-1]
    cmd.run([cmd.name, "/x", "me"])
    assert "'0xa'" == msgs[-1]
    cmd.run([cmd.name, "/o", "me"])
    assert "'0o12'" == msgs[-1]
    return



"""
Functional test of debugger "step" command.
"""
import unittest
from test.functional.fn_helper import compare_output, strarray_setup

# import tracer


class TestStep(unittest.TestCase):
    print("test ", __file__, "skipped")

    def test_step_same_level(self):
        # See that we can step with parameter which is the same as 'step 1'
        cmds = ["step", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5  # NOQA
        y = 6  # NOQA
        ##############################
        d.core.stop()
        out = ["-- x = 5  # NOQA", "-- y = 6  # NOQA"]
        compare_output(self, out, d)
        return

    def test_step_computed_value(self):
        # See that we can step with a computed count value
        cmds = ["step 5-3", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5
        y = 6
        z = 7
        ##############################
        d.core.stop(options={"remove": True})
        out = ["-- x = 5", "-- z = 7"]
        compare_output(self, out, d)

        # Test step>
        cmds = ["step>", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5

        def foo():
            return

        y = 6  # NOQA
        foo()
        ##############################
        d.core.stop(options={"remove": True})
        out = ["-- x = 5", "-> def foo():"]
        compare_output(self, out, d)

        # Test step!
        cmds = ["step!", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        x = 5
        try:
            y = 2
            z = 1 / 0
        except:
            pass
        ##############################
        d.core.stop(options={"remove": True})
        out = ["-- x = 5", "!! z = 1 / 0"]
        compare_output(self, out, d)

        # Test "step" with sets of events. Part 1
        cmds = ["step call exception", "step call exception", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        ##############################
        # x = 5  # NOQA
        # try:

        #     def foo1():
        #         y = 2  # NOQA
        #         raise Exception
        #         return

        #     foo1()
        # except:
        #     pass
        # z = 1  # NOQA
        # ##############################
        # d.core.stop(options={'remove': True})
        # out = ['-- x = 5  # NOQA',
        #        '-> def foo1():',
        #        '!! raise Exception']
        # compare_output(self, out, d)

        # # Test "step" will sets of events. Part 2
        # cmds = ['step call exception 1+0',
        #         'step call exception 1', 'continue']
        # d = strarray_setup(cmds)
        # d.core.start()
        # ##############################
        # x = 5
        # try:
        #     def foo2():
        #         y = 2
        #         raise Exception
        #         return
        #     foo2()
        # except:
        #     pass
        # z = 1
        # ##############################
        # d.core.stop(options={'remove': True})
        # out = ['-- x = 5',
        #        '-> def foo2():',
        #        '!! raise Exception']
        # compare_output(self, out, d)

        return

    # def test_step_between_fn(self):
    #     # Step into and out of a function
    #     def sqr(x):
    #         return x * x

    #     for cmds, out, eventset in (
    #         (
    #             ["step", "step", "continue"],
    #             ["-- x = sqr(4)  # NOQA", "-- return x * x", "-- y = 5  # NOQA"],
    #             frozenset(("line",)),
    #         ),
    #         (
    #             ["step", "step", "step", "step", "continue"],
    #             [
    #                 "-- x = sqr(4)  # NOQA",
    #                 "-> def sqr(x):",
    #                 "-- return x * x",
    #                 "<- return x * x",
    #                 "-- y = 5  # NOQA",
    #             ],
    #             tracer.ALL_EVENTS,
    #         ),
    #     ):
    #         d = strarray_setup(cmds)
    #         d.settings["events"] = eventset
    #         d.core.start()
    #         ##############################
    #         x = sqr(4)  # NOQA
    #         y = 5  # NOQA
    #         ##############################
    #         d.core.stop(options={"remove": True})
    #         compare_output(self, out, d)
    #         pass
    #     return

    def test_step_in_exception(self):
        return

        def boom(x):
            y = 0 / x  # NOQA
            return

        def bad(x):
            boom(x)
            return x * x

        cmds = [
            "step",
            "step",
            "step",
            "step",
            "step",
            "step",
            "step",
            "step",
            "step",
            "step",
            "continue",
        ]
        d = strarray_setup(cmds)
        try:
            d.core.start()
            x = bad(0)
            self.assertTrue(False, "should have raised an exception")
        except ZeroDivisionError:
            self.assertTrue(True, "Got the exception")
        finally:
            d.core.stop(options={"remove": True})
            pass

        out = [
            "-- x = bad(0)  # NOQA",  # line event
            "-> def bad(x):",  # call event
            "-- boom(x)",  # line event
            "-> def boom(x):",  # call event
            "-- y = 0/x  # NOQA",  # line event
            "!! y = 0/x  # NOQA",  # exception event
            "<- y = 0/x  # NOQA",  # return event
            "!! boom(x)",  # exception event
            "<- boom(x)",  # return event
            "!! x = bad(0)  # NOQA",  # exception event
            "-- except ZeroDivisionError:",
        ]
        compare_output(self, out, d)
        return

    pass



"""Unit test for trepan.lib.printing"""

import trepan.lib.printing as printing


def test_lib_printf():
    assert "0o37" == printing.printf(31, "/o")
    assert "00011111" == printing.printf(31, "/t")
    assert "!" == printing.printf(33, "/c")
    assert "0x21" == printing.printf(33, "/x")
    return


def test_lib_argspec():
    assert "test_lib_argspec()" == printing.print_argspec(
        test_lib_argspec, "test_lib_argspec"
    )
    assert printing.print_argspec(None, "invalid_function") is None
    assert (
        printing.print_argspec(printing.print_obj, "printing.print_obj")
        == "printing.print_obj(arg, frame, format=None, short=False) -> str"
    )
    return



# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class ShowDbgTrepan(Mbase_subcmd.DebuggerShowBoolSubcommand):
    """**show dbg_trepan**

    Show debugging the debugger

    See also:
    ---------

    `set dbg_trepan`"""

    min_abbrev = 4  # Min 'show trep"
    short_help = "Show debugging the debugger"
    pass



"""Unit test for trepan.processor.command.quit"""

import threading
from test.unit.cmdhelper import readline_no, readline_yes, setup_unit_test_debugger

import pytest

from trepan.exception import DebuggerQuit
from trepan.processor.command.quit import QuitCommand


def test_quit():
    """Test processor.command.quit.QuitCommand.run()"""
    _, cp = setup_unit_test_debugger()

    #
    # Test that quit without threading raises DebuggerQuit exception
    #

    # First try with  "yes" prompt.
    command = QuitCommand(cp)
    cp.intf[-1].input.readline = readline_yes
    for cmd in ("quit", "q"):
        with pytest.raises(DebuggerQuit):
            command.run([cmd])

    # Now try with  "no" prompt using quit!
    cp.intf[-1].input.readline = readline_no

    with pytest.raises(DebuggerQuit):
        command.run(["quit!"])

    # Test that unconfirmed quit does *not* raise a DebuggerQuit exception
    assert command.run(["quit"]) is None

    # Test that unconfirmed quit does *not* raise a DebuggerQuit exception

    #
    # Test that quit inside threads
    #

    class MyThreadUnconfirmed(threading.Thread):
        def run(self):
            assert command.run(["quit"]) is None
            return

        pass

    t = MyThreadUnconfirmed()
    t.start()
    t.join()

    class MyThreadConfirmed(threading.Thread):
        def run(self):
            with pytest.raises(DebuggerQuit):
                command.run(["quit!"])

            with pytest.raises(DebuggerQuit):
                command.run(["quit!"])

            cp.intf[-1].input.readline = readline_yes
            with pytest.raises(DebuggerQuit):
                assert command.run(["quit"])
            return

        pass

    t = MyThreadConfirmed()
    t.start()
    t.join()

    pass



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2012-2017, 2020-2021, 2023-2024
#  Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect
import linecache
import os.path as osp
import sys
from trepan.processor.cmd_addrlist import INVALID_PARSE_LIST

import pyficache

# Our local modules
from pygments.console import colorize

from trepan.processor.cmdlist import parse_list_cmd

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand

try:
    from trepan.lib.deparse import deparse_cache

    have_deparse = True
except ImportError:
    have_deparse = False


class ListCommand(DebuggerCommand):
    """**list** [ *range* ]

    **list**  **+** | **-** | **.**

    List source code. See `help syntax range` for what can go in a list range.

    Without arguments, print lines starting from where the last list left off
    since the last entry to the debugger. We start off at the location indicated
    by the current stack.

    in addition you can also use:

      - a '.' for the location of the current frame
      - a '-' for the lines before the last list
      - a '+' for the lines after the last list

    Examples:
    --------

        list 5               # List starting from line 5 of current file
        list 5 ,             # Same as above.
        list , 5             # list listsize lines before and up to 5
        list foo.py:5        # List starting from line 5 of file foo.py
        list foo()           # List starting from function foo
        list os.path:5       # List starting from line 5 of module os.path
        list os.path:5, 6    # list lines 5 and 6 of os.path
        list os.path:5, +1   #  Same as above. +1 is an offset
        list os.path:5, 1    # Same as above, since 1 < 5.
        list os.path:5, +6   # list lines 5-11
        list .               # List lines centered from where we currently are stopped
        list -               # List lines previous to those just shown
        list                 # List continuing from where we last left off

    See also:
    ---------

    `set listize` or `show listsize` to see or set the value; `help syntax location`
    for the specification of a location and `help syntax range` for the specification
    of a range."""

    aliases = ("l",)
    short_help = "List source code"

    DebuggerCommand.setup(locals(), category="files", max_args=3)

    def run(self, args):
        proc = self.proc
        dbg_obj = self.core.debugger
        listsize = dbg_obj.settings["listsize"]
        filename, first, last = parse_list_cmd(proc, args, listsize)
        if (filename, first, last) == INVALID_PARSE_LIST:
            return
        curframe = proc.curframe
        if filename is None:
            return
        filename = pyficache.unmap_file(pyficache.resolve_name_to_path(filename))

        # We now have range information. Do the listing.
        max_line = pyficache.size(filename)
        if max_line is None:
            if have_deparse:
                bytecode = curframe.f_code

                if bytecode not in deparse_cache.keys():
                    self.errmsg(
                        'No file %s found; using "deparse" command instead to show source'
                        % filename
                    )
                proc.commands["deparse"].run(["deparse"])
                return
            else:
                return

        canonic_filename = osp.realpath(osp.normcase(filename))

        if first > max_line:
            self.errmsg(
                'Bad start line %d - file "%s" has only %d lines'
                % (first, filename, max_line)
            )
            return

        if last > max_line:
            self.msg("End position changed to last line %d " % max_line)
            last = max_line

        bplist = self.core.bpmgr.bplist
        opts = {
            "reload_on_change": self.settings["reload"],
            "output": self.settings["highlight"],
            "strip_nl": False,
        }

        for field in ("highlight", "style"):
            if field in self.settings:
                opts[field] = self.settings[field]

        if first <= 0:
            first = 1
        try:
            for lineno in range(first, last + 1):
                line = pyficache.getline(filename, lineno, opts)
                if line is None:
                    line = linecache.getline(filename, lineno, proc.frame.f_globals)
                    pass
                if line is None:
                    self.msg("[EOF]")
                    break
                else:
                    line = line.rstrip("\n")
                    s = proc._saferepr(lineno).rjust(3)
                    if len(s) < 5:
                        s += " "
                    if (
                        canonic_filename,
                        lineno,
                    ) in list(bplist.keys()):
                        bp = bplist[
                            (
                                canonic_filename,
                                lineno,
                            )
                        ][0]
                        a_pad = "%02d" % bp.number
                        s += bp.icon_char()
                    else:
                        s += " "
                        a_pad = "  "
                        pass
                    if curframe and lineno == inspect.getlineno(curframe):
                        s += "->"
                        if "plain" != self.settings["highlight"]:
                            s = colorize("bold", s)
                    else:
                        s += a_pad
                        pass
                    self.msg(s + "\t" + line)
                    proc.list_lineno = lineno
                    pass
                pass
            pass
        except KeyboardInterrupt:
            pass
        return False


if __name__ == "__main__":
    # FIXME: make sure the below is in a unit test
    def doit(cmd, args):
        proc = cmd.proc
        proc.current_command = " ".join(args)
        cmd.run(args)

    from trepan.processor.cmdproc import CommandProcessor
    from trepan.processor.command.mock import MockDebugger

    d = MockDebugger()
    cmdproc = CommandProcessor(d.core)
    cmdproc.frame = sys._getframe()
    cmdproc.setup()
    lcmd = ListCommand(cmdproc)

    print("--" * 10)
    # doit(lcmd, ['list'])

    doit(lcmd, ['list', __file__ + ':5'])
    print('--' * 10)

    # doit(lcmd, ['list', 'os:10'])
    # print('--' * 10)

    # doit(lcmd, ['list', '.'])
    # print('--' * 10)

    # doit(lcmd, ['list', '10'])
    # print('--' * 10)

    # doit(lcmd, ['list', '1000'])

    def foo():
        return "bar"

    # doit(lcmd, ['list', 'foo()'])
    # print('--' * 10)
    # doit(lcmd, ['list'])
    # print('--' * 10)
    # doit(lcmd, ['list', '-'])
    # doit(lcmd, ['list', '-'])
    # doit(lcmd, ['list', '+'])
    # doit(lcmd, ['list', '+'])
    doit(lcmd, ["list", "40,", "60"])
    # doit(lcmd, ['list', '20', '5'])

    # doit(lcmd, ['list', 'os.path'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '15'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '30', '3'])
    # print('--' * 10)
    # doit(lcmd, ['list', 'os.path', '40', '50'])
    # print('--' * 10)

    # doit(lcmd, ['list', os.path.abspath(__file__)+':3', '4'])
    # print('--' * 10)
    # doit(lcmd, ['list', os.path.abspath(__file__)+':3', '12-10'])
    # doit(lcmd, ['list', 'os.path:5'])
    pass



"""Unit test for trepan.lib.complete"""

from trepan.lib.breakpoint import BreakpointManager
from trepan.lib.complete import (
    complete_brkpts,
    complete_token,
    complete_token_with_next,
    next_token,
)


def test_complete():
    hash = {"ab": 1, "aac": 2, "aa": 3, "a": 4}
    ary = sorted(hash.keys())
    for result, prefix in [
        [[], "b"],
        [ary, "a"],
        [["aa", "aac"], "aa"],
        [ary, ""],
        [["ab"], "ab"],
        [[], "abc"],
    ]:
        assert result == complete_token(
            ary, prefix
        ), f"Trouble matching {repr(ary)} on {prefix}"
        pass

    for result_keys, prefix in [
        [ary, "a"],
        [["aa", "aac"], "aa"],
        [["ab"], "ab"],
        [[], "abc"],
    ]:
        result = [[key, hash[key]] for key in result_keys]
        assert result == complete_token_with_next(
            hash, prefix
        ), f"Trouble matching {repr(hash)} on {prefix}"
        pass

    return


def test_next_token():
    x = "  now is  the  time"
    for pos, expect in [
        [0, [5, "now"]],
        [2, [5, "now"]],
        [5, [8, "is"]],
        [8, [13, "the"]],
        [9, [13, "the"]],
        [13, [19, "time"]],
        [19, [19, ""]],
    ]:
        assert expect == next_token(x, pos), f"Trouble with next_token({x}, {pos})"
        pass

    return


def test_complete_brkpts():
    bpmgr = BreakpointManager()
    bp = bpmgr.add_breakpoint("foo", 10, 5)
    assert bp
    for find in "1":
        assert complete_brkpts(bpmgr, find) == [
            "1"
        ], f"breakpoint completion of '{find}'"
    for find in ("2", "11"):
        assert (
            complete_brkpts(bpmgr, find) == []
        ), f"breakpoint non-completion of '{find}'"
        pass
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2010, 2012-2013, 2015, 2020, 2023 Rocky
#   Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
from pygments.console import colorize

NotImplementedMessage = "This method must be overridden in a subclass"

__all__ = ["Processor"]


class Processor(object):
    """A processor is the thing that handles the events that come to
    the debugger.  It has it's own I/O mechanism and a way to handle
    the events.
    """

    def __init__(self, core_obj):
        self.core = core_obj
        self.debugger = core_obj.debugger
        return

    # Note for errmsg, msg, and msg_nocr we don't want to simply make
    # an assignment of method names like self.msg = self.intf.msg,
    # because we want to allow the interface (intf) to change
    # dynamically. That is, the value of self.debugger may change
    # in the course of the program and if we made such an method assignment
    # we wouldn't pick up that change in our self.msg
    def errmsg(self, message, opts={}):
        """Convenience short-hand for self.intf[-1].errmsg"""
        if "plain" != self.debugger.settings["highlight"]:
            message = colorize("standout", message)
            pass
        return self.intf[-1].errmsg(message)

    def msg(self, msg, opts={}):
        """Convenience short-hand for self.debugger.intf[-1].msg"""
        return self.intf[-1].msg(msg)

    def msg_nocr(self, msg, opts={}):
        """Convenience short-hand for self.debugger.intf[-1].msg_nocr"""
        return self.intf[-1].msg_nocr(msg)

    def event_processor(self, frame, event, arg):
        raise NotImplementedError(NotImplementedMessage)

    def rst_msg(self, text, opts={}):
        """Convert ReStructuredText and run through msg()"""
        from trepan.lib.format import rst_text

        text = rst_text(
            text,
            "plain" == self.debugger.settings["highlight"],
            self.debugger.settings["width"],
        )
        return self.msg(text)

    def section(self, message, opts={}):
        if "plain" != self.settings("highlight"):
            message = colorize("bold", message)
            pass
        return self.msg(message, opts)

    def settings(self, setting):
        return self.core.debugger.settings[setting]

    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009-2010, 2013-2015, 2017, 2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger input possibly attached to a user or interactive. """

import io
import sys

from trepan import misc as Mmisc
from trepan.inout import base as Mbase


def readline_importable() -> bool:
    try:
        import readline  # NOQA

        return True
    except ImportError:
        return False
    return  # Not reached


class DebuggerUserInput(Mbase.DebuggerInputBase):
    """Debugger input connected to what we think of as a end-user input
    as opposed to a relay mechanism to another process. Input could be
    interactive terminal, but it might be file input."""

    def __init__(self, inp=None, opts=None):
        self.input = inp or sys.stdin
        self.line_edit = None  # Our name for GNU readline capability
        self.open(self.input, opts)
        return

    def close(self):
        self.input.close()
        self.closed = True
        return

    DEFAULT_OPEN_READ_OPTS = {
        "use_raw": True,
        "try_readline": True,
    }

    def use_history(self):
        return self.use_raw and readline_importable()

    def open(self, inp, opts={}):
        """Use this to set where to read from.

        Set opts['try_lineedit'] if you want this input to interact
        with GNU-like readline library. By default, we will assume to
        try importing and using readline. If readline is not
        importable, line editing is not available whether or not
        opts['try_readline'] is set.

        Set opts['use_raw'] if input should use Python's use_raw(). If
        however 'inp' is a string and opts['use_raw'] is not set, we
        will assume no raw output. Note that an individual readline
        may override the setting.
        """
        get_option = lambda key: Mmisc.option_set(
            opts, key, self.DEFAULT_OPEN_READ_OPTS
        )
        if (
            isinstance(inp, io.TextIOWrapper)
            or isinstance(inp, io.StringIO)
            or hasattr(inp, "isatty")
            and inp.isatty()
        ):
            self.use_raw = get_option("use_raw")
        elif isinstance(inp, "string".__class__):  # FIXME
            if opts is None:
                self.use_raw = False
            else:
                self.use_raw = get_option("use_raw")
                pass
            inp = open(inp, "r")
        else:
            raise IOError("Invalid input type (%s) for %s" % (type(inp), inp))
        self.input = inp
        self.line_edit = get_option("try_readline") and readline_importable()
        self.closed = False
        return

    def readline(self, use_raw=None, prompt=""):
        """Read a line of input. EOFError will be raised on EOF.

        Note: some user interfaces may decide to arrange to call
        DebuggerOutput.write() first with the prompt rather than pass
        it here.. If `use_raw' is set raw_input() will be used in that
        is supported by the specific input input. If this option is
        left None as is normally expected the value from the class
        initialization is used.
        """
        # FIXME we don't do command completion.
        if use_raw is None:
            use_raw = self.use_raw
            pass
        if use_raw:
            try:
                inp = input(prompt)
                # import pdb; pdb.set_trace()
                return inp
            except ValueError:
                raise EOFError
            pass

        else:
            line = self.input.readline()
            if not line:
                raise EOFError
            return line.rstrip("\n")
        pass

    pass


# Demo
if __name__ == "__main__":
    print("readline importable: ", readline_importable())
    inp = DebuggerUserInput(__file__)
    line = inp.readline()
    print(line)
    inp.close()
    inp.open("input.py", opts={"use_raw": False})
    while True:
        try:
            inp.readline()
        except EOFError:
            break
        pass
    try:
        inp.readline()
    except EOFError:
        print("EOF handled correctly")
    pass

    if len(sys.argv) > 1:
        inp = DebuggerUserInput()
        try:
            print("Type some characters:", end=" ")
            line = inp.readline()
            print("You typed: %s" % line)
            print("Type some more characters (raw):", end=" ")
            line = inp.readline(True)
            print("Type even more characters (not raw):", end=" ")
            line = inp.readline(True)
            print("You also typed: %s" % line)
        except EOFError:
            print("Got EOF")
        pass
    pass



#!/usr/bin/env python3
"""Unit test for trepan.processor.cmdproc.break"""
import inspect
import os.path as osp
from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.cmdbreak import parse_break_cmd


def canonic_tuple(t):
    fname = t[1]
    if fname:
        if fname.endswith(".pyc"):
            fname = fname[:-1]
        got = list(t)
        got[1] = osp.basename(fname)
        t = tuple(got)
    return t


def test_cmd_break():
    d, cp = setup_unit_test_debugger()
    for expect, cmd in (
        ((None, None, None), "break '''c:\\tmp\\foo.bat''':1"),
        ((None, None, None), 'break """/Users/My Documents/foo.py""":2'),
        (("<module>", osp.basename(__file__), 10), "break 10"),
        ((None, None, None), "break cmdproc.py:5"),
        ((None, None, None), "break set_break()"),
        (("<module>", osp.basename(__file__), 4, "i==5"), "break 4 if i==5"),
        ((None, None, None), "break cmdproc.setup()"),
    ):
        args = cmd.split(" ")
        cp.current_command = cmd
        got = canonic_tuple(parse_break_cmd(cp, args))
        assert expect == tuple(got[: len(expect)]), cmd
        # print(got)

    cp.frame = inspect.currentframe()
    cp.setup()

    # WARNING: magic number after f_lineno is fragile on the number of tests!
    break_lineno = cp.frame.f_lineno + 7
    for expect, cmd in (
        ((None, osp.basename(__file__), break_lineno, None), "break"),
        ((None, osp.basename(__file__), break_lineno, "True"), "break if True"),
    ):
        args = cmd.split(" ")
        cp.current_command = cmd
        got = canonic_tuple(parse_break_cmd(cp, args))
        print(parse_break_cmd(cp, args))
        assert expect == got[: len(expect)], cmd

    print(break_lineno)
    pass
    return



"""
Functional test of debugger signal handling.
"""
import os
import signal
import unittest
from test.functional.fn_helper import compare_output, strarray_setup

import pytest


class TestSigHandler(unittest.TestCase):
    @pytest.mark.skip("This does not work when run with other pytests")
    def test_handle(self):
        # See that we handle a USR1 signal with a 'stop' action
        cmds = ["handle usr1 stop nopass", "continue", "where", "continue"]
        d = strarray_setup(cmds)

        def signal_handler(num, _):
            print(f"signal {num} received")
            return

        signal.signal(signal.SIGUSR1, signal_handler)
        d.core.start()
        ##############################
        x = 5
        os.kill(os.getpid(), signal.SIGUSR1)
        y = 6  # +1
        ##############################        # +2
        d.core.stop()  # +3
        out = [
            "-- x = 5",
            ("?! os.kill(os.getpid(), signal.SIGUSR1)"),
        ]
        compare_output(self, out, d)

        # Now define a signal handler and see that this is skill okay
        # "ignore" is the same as "nopass"
        # FIXME: add this common code to the helper.
        cmds = ["handle usr1 stop ignore", "continue", "where", "continue"]
        d.intf[-1].output.output = [""]
        d.core.step_ignore = 0
        d.intf[-1].input.input = cmds

        def signal_handler2(num, f):
            print("signal %d received" % num)
            return

        signal.signal(signal.SIGUSR1, signal_handler2)

        d.core.start()
        ##############################
        x = 7
        os.kill(os.getpid(), signal.SIGUSR1)
        y = 8  # +1
        ##############################        # +2
        d.core.stop()  # +3
        out = [
            "-- x = 7",
            ("?! os.kill(os.getpid(), signal.SIGUSR1)"),
        ]
        compare_output(self, out, d)

        # How about USR2 signal with 'ignore' and 'noprint' actions?
        cmds = [
            "handle usr2 ignore nostop noprint",
            "continue",
            "info signal usr2",
            "continue",
        ]
        d.intf[-1].output.output = [""]
        d.core.step_ignore = 0
        d.intf[-1].input.input = cmds

        def signal_handler3(num, f):
            print(f"signal {num} received")
            return

        signal.signal(signal.SIGUSR2, signal_handler2)

        d.core.start()
        ##############################
        x = 9  # NOQA
        os.kill(os.getpid(), signal.SIGUSR2)
        y = 10  # NOQA
        ##############################
        d.core.stop()
        out = ["-- x = 9  # NOQA"]
        compare_output(self, out, d)

        return

    pass



"""
Functional test of debugger "jump" command.
"""
import inspect
import os
import unittest
from test.functional.fn_helper import compare_output, strarray_setup


class TestJump(unittest.TestCase):
    @unittest.skipIf(
        "TRAVIS" in os.environ, "FIXME: figure out why this doesn't work in travis"
    )
    def test_skip(self):
        # See that we can jump with line number
        curframe = inspect.currentframe()
        cmds = ["step", "jump %d" % (curframe.f_lineno + 7), "continue"]
        d = strarray_setup(cmds)  # 1
        d.core.start()  # 2
        ##############################          # 3...
        x = 4
        x = 5
        x = 6
        z = 7  # NOQA
        ##############################
        d.core.stop(options={"remove": True})
        out = [
            "-- x = 4",  # x = 4 is shown in prompt, but not run.
            "-- x = 5",
            "-- z = 7  # NOQA",
        ]
        compare_output(self, out, d)
        self.assertEqual(4, x)  # Make sure x = 5 and 6 were skipped.
        return

    pass


if __name__ == "__main__":
    unittest.main()



"""Unit test for trepan.processor.command.help"""

from test.unit.cmdhelper import setup_unit_test_debugger

from trepan.processor.command.help import HelpCommand, categories

errors = []
msgs = []
d, cp = setup_unit_test_debugger()
cmd = HelpCommand(cp)


def errmsg(msg_str: str):
    errors.append(msg_str)
    return


def msg(msg_str: str):
    msgs.append(msg_str)
    return


cmd.msg = msg
cmd.errmsg = errmsg


def reset():
    global errors
    errors = []

    global msgs
    msgs = []


def test_help_command():
    """Test we can run 'help *cmd* for each command"""

    reset()
    for name in cp.commands.keys():
        cmd.run(["help", name])
        pass

    assert len(msgs) > 0, "Should get help output"
    assert 0 == len(errors), "Should not get errors"
    return


def test_help_categories():
    """Test we can run 'help *cmd* for each category"""

    reset()

    for name in categories.keys():
        cmd.run(["help", name])
        pass

    assert len(msgs) > 0, "Should get help output"
    assert 0 == len(errors), "Should not get errors"

    for name in categories.keys():
        cmd.run(["help", name, "*"])
        pass
    assert len(msgs) > 0, "Should get help output"
    assert 0 == len(errors), "Should not get errors"
    return


def test_short_help():
    """Test each command has some sort of short help"""
    for cmd in list(cp.commands.values()):
        assert str == type(cmd.short_help)
        pass
    return



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2023 Rocky Bernstein <rocky@gnu.org>
""" Common I/O routines"""


# Note for errmsg, msg, and msg_nocr we don't want to simply make
# an assignment of method names like self.msg = self.debugger.intf.msg,
# because we want to allow the interface (intf) to change
# dynamically. That is, the value of self.debugger may change
# in the course of the program and if we made such an method assignment
# we wouldn't pick up that change in our self.msg
def errmsg(proc_obj, message, opts={}):
    response = proc_obj.response
    if "set_name" in opts:
        response["name"] = "error"
    return response["errs"].append(message)


def msg(proc_obj, message, opts={}):
    response = proc_obj.response
    return response["msg"].append(message)


# Demo it
if __name__ == "__main__":

    class Demo:
        def __init__(self):
            self.response = {"errs": [], "msg": []}
            pass

        pass

    import pprint

    demo = Demo()
    msg(demo, "hi")
    pp = pprint.PrettyPrinter()
    pp.pprint(demo.response)



"""Unit test for trepan.inout.input"""
import os.path as osp

from trepan.inout.input import DebuggerUserInput

srcdir = osp.abspath(osp.dirname(__file__))


def test_DebuggerInput():
    cmdhelper_file = osp.join(srcdir, "..", "cmdhelper.py")
    inp = DebuggerUserInput(cmdhelper_file)
    assert isinstance(
        inp, DebuggerUserInput
    ), "Should have gotten a DebuggerUserInput object back"
    line = inp.readline()
    assert "# -*- coding: utf-8 -*-" == line
    inp.close()
    # Should be okay
    inp.close()
    return



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2013-2015, 2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Debugger FIFO Input/Output interface. """

import os
import tempfile

from trepan import misc as Mmisc
from trepan.inout.base import DebuggerInOutBase
from trepan.lib import default as Mdefault, file as Mfile


class FIFOClient(DebuggerInOutBase):
    """Debugger Client Input/Output Socket."""

    DEFAULT_INIT_OPTS = {"open": True}

    def __init__(self, inp=None, opts=None):
        get_option = lambda key: Mmisc.option_set(
            opts, key, Mdefault.CLIENT_SOCKET_OPTS
        )
        self.state = "disconnected"
        self.flush_after_write = True
        self.input = None
        self.output = None
        self.line_edit = False  # Our name for GNU readline capability
        self.state = "disconnected"
        open_pid = get_option("open")
        if open_pid:
            self.open(open_pid)
            pass
        return

    def close(self):
        """Closes both input and output"""
        self.state = "closing"
        if self.input:
            self.input.close()
            pass
        if self.output:
            self.output.close()
            pass
        self.state = "disconnnected"
        return

    def flush(self):
        return self.output.flush()

    def open(self, pid, opts=None):

        # Not in/out are reversed from server side
        d = tempfile.gettempdir()
        self.out_name = os.path.join(d, ("trepan-%s.in" % pid))
        self.in_name = os.path.join(d, ("trepan-%s.out" % pid))
        is_readable = Mfile.readable(self.out_name)
        if not is_readable:
            if is_readable is None:
                raise IOError("output FIFO %s doesn't exist" % self.out_name)
            else:
                raise IOError("output FIFO %s is not readable" % self.out_name)
            pass
        is_readable = Mfile.readable(self.in_name)
        if not is_readable:
            if is_readable is None:
                raise IOError("input FIFO %s doesn't exist" % self.in_name)
            else:
                raise IOError("input FIFO %s is not readable" % self.out_name)
            pass
        self.state = "active"
        return

    def read_msg(self):
        """Read a line of input. EOFError will be raised on EOF.

        Note that we don't support prompting"""
        # FIXME: do we have to create and check a buffer for
        # lines?
        if self.state == "active":
            if not self.input:
                self.input = open(self.in_name, "r")
                pass
            line = self.input.readline()
            if not line:
                self.state = "disconnected"
                raise EOFError
            return line.encode("utf-8")
        else:
            raise IOError("readline called in state: %s." % self.state)
        return  # Not reached

    def write(self, msg):
        """This method the debugger uses to write. In contrast to
        writeline, no newline is added to the end to `str'.
        """
        if self.state == "active":
            if not self.output:
                self.output = open(self.out_name, "w")
                pass
            pass
        else:
            raise EOFError
        self.output.write(msg)
        if self.flush_after_write:
            self.flush()
        return


# Demo
if __name__ == "__main__":
    fifo = FIFOClient(opts={"open": False})
    import sys

    if len(sys.argv) > 1:
        print(
            "Connecting...",
        )
        fifo.open(sys.argv[1])
        print("connected.")
        while True:
            prompt = fifo.readline()
            line = input(prompt)
            if len(line) == 0:
                break
            try:
                line = fifo.writeline(line)
                print("Got: ", fifo.readline())
            except EOFError:
                break
            pass
        pass
    fifo.close()
    pass



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2013-2014, 2017, 2023 Rocky Bernstein
#   <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Module for client (i.e. user to communication-device) interaction.
The debugged program is at the other end of the communication."""

from trepan.inout import fifoclient as Mfifoclient, tcpclient as Mtcpclient

# Our local modules
from trepan.interfaces import user as Muser

DEFAULT_INIT_CONNECTION_OPTS = {"IO": "TCP"}


class ClientInterface(Muser.UserInterface):
    """Interface for a user which is attached to a debugged process
    via some sort of communication medium (e.g. socket, tty, FIFOs).
    This could be on the same computer in a different process or on
    a remote computer."""

    def __init__(
        self, inp=None, out=None, inout=None, user_opts={}, connection_opts={}
    ):

        opts = DEFAULT_INIT_CONNECTION_OPTS.copy()
        opts.update(connection_opts)

        Muser.UserInterface.__init__(self, inp, out, user_opts)

        self.inout = None  # initialize in case assignment below fails
        if inout:
            self.inout = inout
        else:
            self.server_type = opts["IO"]
            if "FIFO" == self.server_type:
                self.inout = Mfifoclient.FIFOClient(opts=opts)
            elif "TCP" == self.server_type:
                self.inout = Mtcpclient.TCPClient(opts=opts)
            else:
                self.errmsg(
                    "Expecting server type TCP or FIFO. Got: %s." % self.server_type
                )
                return
            pass
        return

    def read_remote(self):
        """Send a message back to the server (in contrast to
        the local user output channel)."""
        coded_line = self.inout.read_msg()
        if isinstance(coded_line, bytes):
            coded_line = coded_line.decode("utf-8")
        control = coded_line[0]
        remote_line = coded_line[1:]
        return (control, remote_line)

    def write_remote(self, code, msg):
        """Send a message back to the server (in contrast to
        the local user output channel)."""
        # FIXME change into write_xxx
        return self.inout.writeline(code + msg)

    pass


# Demo
if __name__ == "__main__":
    intf = ClientInterface()



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2012-2013 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.processor import cmdfns as Mcmdfns


class SetCmdtrace(Mbase_subcmd.DebuggerSetBoolSubcommand):
    """Set echoing lines read from debugger command files"""

    in_list = True
    min_abbrev = len("cmdt")  # Need at least "set cmdt"
    pass

    def run(self, args):
        Mcmdfns.run_set_bool(self, args)
        dbg = self.debugger
        if hasattr(dbg.intf[-1], "verbose"):
            dbg.intf[-1].verbose = dbg.settings[self.name]
            pass
        return

    pass



"""
Functional test of debugger "next" command.
"""
import unittest
from test.functional.fn_helper import compare_output, strarray_setup


class TestNext(unittest.TestCase):
    print("test ", __file__, "skipped")

    def test_next_same_level(self):
        # See that we can next with parameter which is the same as 'next 1'
        cmds = ["next", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        x = 5
        y = 6
        d.core.stop()
        out = ["-- x = 5", "-- y = 6"]
        compare_output(self, out, d)

        # See that we can next with a computed count value
        cmds = ["next 5-3", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        x = 5  # NOQA
        y = 6  # NOQA
        z = 7  # NOQA
        d.core.stop(options={"remove": True})
        out = ["-- x = 5  # NOQA", "-- z = 7  # NOQA"]
        compare_output(self, out, d)
        return

    def test_next_between_fn(self):
        # Next over a function
        def fact(x):
            if x <= 1:
                return 1
            return fact(x - 1)

        cmds = ["next", "continue"]
        d = strarray_setup(cmds)
        d.core.start()
        x = fact(4)  # NOQA
        y = 5  # NOQA
        d.core.stop(options={"remove": True})
        out = ["-- x = fact(4)  # NOQA", "-- y = 5  # NOQA"]
        compare_output(self, out, d)
        return

    def test_next_in_exception(self):
        return

        def boom(x):
            y = 0 / x  # NOQA
            return

        def buggy_fact(x):
            if x <= 1:
                return boom(0)
            return buggy_fact(x - 1)

        cmds = ["next", "continue"]
        d = strarray_setup(cmds)
        try:
            d.core.start()
            x = buggy_fact(4)  # NOQA
            y = 5  # NOQA
            self.assertTrue(False, "should have raised an exception")
        except ZeroDivisionError:
            self.assertTrue(True, "Got the exception")
        finally:
            d.core.stop(options={"remove": True})
            pass

        out = ["-- x = buggy_fact(4)  # NOQA", "!! x = buggy_fact(4)  # NOQA"]
        compare_output(self, out, d, cmds)
        return

    pass


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2024 Rocky Bernstein
#

from pygments.styles import STYLE_MAP

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand

style_names = sorted(list(STYLE_MAP.keys()))


class ShowStyles(DebuggerSubcommand):
    """**show styles**

    Give a list of all pygments style available in formatting 256-color terminal text.

    See also:
    ---------

    `show style`, `set style`, `show highlight`"""

    in_list = True
    short_help = "Show all pygments style available"
    min_abbrev = len("styles")

    def run(self, args):
        if len(args) != 0:
            self.errmsg("Expecting no args")
            return

        self.section("style names: ")
        self.msg(self.columnize_commands(style_names))
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    sub = Mhelper.demo_run(ShowStyle, [])
    d = sub.proc.debugger
    sub.run(["show"])
    pass



# -*- coding: utf-8 -*-
#
#  Copyright (C) 2009, 2013, 2015, 2020, 2023 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
import inspect
import os

from trepan.processor import cmdproc as Mcmdproc

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class JumpCommand(DebuggerCommand):
    """**jump** *lineno*

    Set the next line that will be executed. The line must be within the
    stopped or bottom-most execution frame."""

    aliases = ("j",)
    category = "running"
    execution_set = ["Running"]
    min_args = 1
    max_args = 1
    name = os.path.basename(__file__).split(".")[0]
    need_stack = False
    short_help = "Set the next line to be executed"

    def run(self, args):
        if not self.core.is_running():
            return False

        if self.proc.curindex + 1 != len(self.proc.stack):
            self.errmsg("You can only jump within the bottom frame")
            return False

        if self.proc.curframe.f_trace is None:
            self.errmsg("Sigh - operation can't be done here.")
            return False

        lineno = self.proc.get_an_int(
            args[1], ("jump: a line number is required, " + "got %s.") % args[1]
        )
        if lineno is None:
            return False
        try:
            # Set to change position, update our copy of the stack,
            # and display the new position
            print(self.proc.curframe.f_trace)
            self.proc.curframe.f_lineno = lineno
            self.proc.stack[self.proc.curindex] = (
                self.proc.stack[self.proc.curindex][0],
                lineno,
            )
            Mcmdproc.print_location(self.proc)
        except ValueError as e:
            self.errmsg("jump failed: %s" % e)
        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock

    d, cp = mock.dbg_setup()
    command = JumpCommand(cp)
    print("jump when not running: ", command.run(["jump", "1"]))
    command.core.execution_status = "Running"
    cp.curframe = inspect.currentframe()
    cp.curindex = 0
    cp.stack = Mcmdproc.get_stack(cp.curframe, None, None)
    command.run(["jump", "1"])
    cp.curindex = len(cp.stack) - 1
    command.run(["jump", "1"])
    pass



#  Copyright (C) 2013, 2015, 2020, 2024 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from trepan.lib.complete import complete_token

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand


class UnaliasCommand(DebuggerCommand):
    """**unalias** *alias-name*

    Remove alias *alias-name*

    See also:
    ---------

    'alias'"""

    min_args = 1
    short_help = "Remove an alias"

    DebuggerCommand.setup(locals(), category="support", max_args=1)

    def complete(self, prefix):
        return complete_token(self.proc.aliases.keys(), prefix)

    # Run command.
    def run(self, args):
        for arg in args[1:]:
            if arg in self.proc.aliases:
                del self.proc.aliases[arg]
                self.msg(f"Alias for {arg} removed.")
            else:
                self.msg(f"No alias found for {arg}")
                pass
            pass
        return

    pass


if __name__ == "__main__":
    # Demo it.
    from trepan import debugger

    d = debugger.Trepan()
    cp = d.core.processor
    command = UnaliasCommand(cp)
    command.run(["unalias", "s"])
    command.run(["unalias", "s"])
    command.run(["unalias", "foo", "n"])
    print(command.complete(""))
    pass



"""Unit test for trepan.misc"""

from trepan.misc import bool2YN, option_set, wrapped_lines


def test_bool2YN():
    for arg, expect in ((True, "Y"), (False, "N"), (None, "N")):
        assert bool2YN(arg) == expect


def test_option_set():
    """Test trepan.misc.option_set()"""
    TEST_OPTS = {"a": True, "b": 5, "c": None}

    def get_option(key):
        return option_set(opts, key, TEST_OPTS)

    opts = {"d": 6, "a": False}
    for opt, expect in [("a", False), ("b", 5), ("c", None), ("d", 6)]:
        assert expect == get_option(opt)
    opts = None
    for opt, expect in [("a", True), ("b", 5), ("c", None), ("d", None)]:
        assert expect == get_option(opt)
    pass


def test_wrapped_msg():
    """Test trepan.misc.wrapped_lines()"""
    assert "hi there" == wrapped_lines("hi", "there", 80)
    assert "hi\n\tthere" == wrapped_lines("hi", "there", 5)



"""Unit test for trepan.processor.command.kill"""

import signal
import sys
from test.unit.cmdhelper import dbg_setup

from trepan.processor.command import kill as Mkill


def test_kill():
    """Test processor.command.kill.KillCommand.run()"""

    signal_caught = False

    def handle(*args):
        signal_caught = True

    return

    signal.signal(28, handle)
    d, cp = dbg_setup()
    command = Mkill.KillCommand(cp)
    result = command.run(["kill", "wrong", "number", "of", "args"])
    assert not result
    assert not signal_caught
    if sys.platform != "win32":
        result = command.run(["kill", "28"])
        assert not result
        assert signal_caught
    return



# -*- coding: utf-8 -*-
#   Copyright (C) 2008-2009, 2013, 2015-2015 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


class InfoArgs(Mbase_subcmd.DebuggerSubcommand):
    """**info args**

    Show parameters of the current stack frame.

    See also:
    ---------
    `info locals`, `info globals`, `info frame`"""

    min_abbrev = 1
    need_stack = True
    short_help = "Argument variables of the current stack frame"

    def run(self, args):
        if not self.proc.curframe:
            self.errmsg("No stack.")
            return False
        f = self.proc.curframe
        co = f.f_code
        # Break out display into args, varargs, keywords, and locals ?
        # args, varargs, varkw, f_locals = getargvalues(f)
        d = f.f_locals
        n = co.co_argcount
        if co.co_flags & inspect.CO_VARARGS:
            n += 1
        if co.co_flags & inspect.CO_VARKEYWORDS:
            n += 1

        if n == 0:
            self.msg("no parameters")
        else:
            self.section("Argument parameters")
            for i in range(n):
                name = co.co_varnames[i]
                self.msg_nocr("%d: %s = " % (i + 1, name))
                if name in d:
                    self.msg(str(d[name]))
                else:
                    self.ermsg("undefined")
                    pass
                pass
            pass
        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock, info as Minfo

    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoArgs(i)
    print(sub.run([]))
    cp.curframe = inspect.currentframe()
    print(sub.run([]))

    def nest_me(sub, cp, b=1):
        cp.curframe = inspect.currentframe()
        print(sub.run([]))
        return

    print("-" * 10)
    nest_me(sub, cp, 3)
    print("-" * 10)
    nest_me(sub, cp)
    pass



"""Unit test for trepan.processor.command.run"""


from test.unit.cmdhelper import readline_yes, setup_unit_test_debugger

import pytest

from trepan.exception import DebuggerRestart
from trepan.processor.command.run import RunCommand


def test_run():
    """Test processor.command.run.RunCommand.run()"""
    _, cp = setup_unit_test_debugger()
    command = RunCommand(cp)

    # First try with  "yes" prompt.
    cp.intf[-1].input.readline = readline_yes

    for cmd in ("run", "R"):
        with pytest.raises(DebuggerRestart):
            command.run([cmd])

    # Now try with  "no" prompt using run!
    def confirm_no(_, default) -> bool:
        return False

    command.confirm = confirm_no

    with pytest.raises(DebuggerRestart):
        command.run(["run!"])

    # Test that unconfirmed quit does *not* raise a DebuggerQuit exception
    assert command.run(["run"]) is None

    return



# -*- coding: utf-8 -*-
#  Copyright (C) 2009, 2013-2015, 2017, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete import complete_bpnumber


class DisableCommand(DebuggerCommand):
    """**disable** *bpnumber* [*bpnumber* ...]

    Disables the breakpoints given as a space separated list of breakpoint
    numbers. To disable all breakpoints, give no argument. See also `info break` to get a list.

    See also:
    ---------
    `enable`"""

    short_help = "Disable some breakpoints"

    complete = complete_bpnumber

    DebuggerCommand.setup(locals(), category="breakpoints")

    def run(self, args):
        if len(args) == 1:
            self.msg(self.core.bpmgr.en_disable_all_breakpoints(do_enable=False))
            return
        #         if args[1] == 'display':
        #             self.display_enable(args[2:], 0)
        #             return
        for i in args[1:]:
            success, msg = self.core.bpmgr.en_disable_breakpoint_by_number(
                int(i), False
            )
            if not success:
                self.errmsg(msg)
            else:
                self.msg("Breakpoint %s disabled." % i)
                pass
            pass
        return


if __name__ == "__main__":
    from trepan.debugger import Trepan

    d = Trepan()
    command = DisableCommand(d.core.processor)
    pass



# This code fragment what is embedded inside ../test_jump.py
# We have it here to make it easy to test things when things go wrong.


# commands are:
#   step
#   jump 13 # 6 + 7
#   continue
##############################  # 3...
x = 4
x = 5
x = 6
z = 7  # NOQA



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2009, 2014, 2023 Rocky Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import sys
import threading

from trepan.lib import stack as Mstack, thred as Mthread

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd


# FIXME turn into yet another subcommand thingy.
class InfoThread(Mbase_subcmd.DebuggerSubcommand):
    """**info threads** [*thread-name*|*thread-number*] [**terse**|**verbose**]

    List all currently-known thread name(s).

    If no thread name is given, we list info for all threads. Unless a
    terse listing, for each thread we give:

      - the class, thread name, and status as *Class(Thread-n, status)*

      - the top-most call-stack information for that thread.

    Generally the top-most calls into the debugger and dispatcher are
    omitted unless set dbg_trepan is *True*.

    If 'verbose' appended to the end of the command, then the entire stack
    trace is given for each frame.  If 'terse' is appended we just list
    the thread name and thread id.

    To get the full stack trace for a specific thread pass in the thread name."""

    min_abbrev = 2  # Min is "info th"
    max_args = 2
    need_stack = True
    short_help = "List thread info"

    def __init__(self, cmd):
        Mbase_subcmd.DebuggerSubcommand.__init__(self, cmd)
        self.name2id = {}
        return

    def stack_trace(self, f):
        """A mini stack trace routine for threads."""
        while f:
            if (
                not self.core.ignore_filter.is_included(f)
                or self.settings["dbg_trepan"]
            ):
                s = Mstack.format_stack_entry(self, (f, f.f_lineno))
                self.msg(" " * 4 + s)
                pass
            f = f.f_back
            pass
        return

    def info_thread_terse(self, name2id, arg=None):
        if arg is not None:
            thread_name = arg
            if thread_name in list(name2id.keys()):
                self.info_thread_line(thread_name, name2id)
            else:
                self.errmsg("Don't know about thread name %s" % thread_name)
                return
            pass

        # Show all threads
        thread_name_list = list(name2id.keys())
        thread_name_list.sort()
        for thread_name in thread_name_list:
            self.info_thread_line(thread_name, name2id)
            pass
        # self.info_thread_missing()
        return

    def info_thread_line(self, thread_name, name2id):
        if thread_name == self.proc.frame_thread_name:
            prefix = "-> "
        elif thread_name == self.proc.thread_name:
            prefix = "=> "
        else:
            prefix = "   "
            pass

        self.msg("%s%s: %d" % (prefix, thread_name, name2id[thread_name]))
        return

    def run(self, args):
        # FIXME: add thread locking here?

        self.thread_name = Mthread.current_thread_name()

        name2id = Mthread.map_thread_names()
        # invert threading._active
        for thread_id in list(threading._active.keys()):
            thread = threading._active[thread_id]
            name = thread.getName()
            if name not in list(self.name2id.keys()):
                self.name2id[name] = thread_id
                pass
            pass

        all_verbose = False
        if len(args) == 1:
            if args[0].startswith("verbose"):
                all_verbose = True
            elif args[0].startswith("terse"):
                self.info_thread_terse(name2id)
                return
            pass

        if len(args) > 0 and not all_verbose:
            thread_name = args[0]
            if thread_name == ".":
                thread_name = self.thread_name
            try:
                thread_id = int(thread_name)
                if thread_id not in list(threading._active.keys()):
                    self.errmsg("Don't know about thread number %s" % thread_name)
                    self.info_thread_terse(name2id)
                    return
            except ValueError:
                if thread_name not in list(self.name2id.keys()):
                    self.errmsg("Don't know about thread %s" % thread_name)
                    self.info_thread_terse(name2id)
                    return
                thread_id = self.name2id[thread_name]
                pass

            frame = sys._current_frames()[thread_id]
            self.stack_trace(frame)
            return

        # Show info about *all* threads
        thread_key_list = list(self.name2id.keys())
        thread_key_list.sort()
        for thread_name in thread_key_list:
            thread_id = self.name2id[thread_name]
            frame = sys._current_frames()[thread_id]
            s = ""
            # Print location where thread was created and line number
            if thread_id in threading._active:
                thread = threading._active[thread_id]
                thread_name = thread.getName()
                if thread_name == self.proc.frame_thread_name:
                    prefix = "-> "
                    if not self.settings["dbg_trepan"]:
                        frame = Mthread.find_debugged_frame(frame)
                        pass
                    pass
                elif thread_name == self.proc.thread_name:
                    prefix = "=> "
                else:
                    prefix = "   "
                    pass
                s += "%s%s" % (prefix, str(thread))
                if all_verbose:
                    s += ": %d" % thread_id
                    pass
            else:
                s += "    thread id: %d" % thread_id
                pass
            s += "\n    "
            s += Mstack.format_stack_entry(
                self, (frame, frame.f_lineno), color=self.settings["highlight"]
            )
            self.section("-" * 40)
            self.msg(s)
            frame = frame.f_back
            if all_verbose and frame:
                self.stack_trace(frame)
                pass
        return

    pass


if __name__ == "__main__":
    from trepan.processor.command import info as Minfo, mock

    d, cp = mock.dbg_setup()
    i = Minfo.InfoCommand(cp)
    sub = InfoThread(i)
    import inspect

    cp.curframe = inspect.currentframe()
    sub.run([])
    print("=" * 30)
    sub.run(["foo"])
    print("=" * 30)
    sub.run(["MainThread"])
    print("=" * 30)
    sub.run(["terse"])
    print("=" * 30)
    sub.run(["verbose"])
    print("=" * 30)
    sub.run(["MainThread", "verbose"])
    # Try with threading.
    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2009, 2013, 2015-2016, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSetBoolSubcommand


class SetDifferent(DebuggerSetBoolSubcommand):
    """**set different** [ **on** | **off** ]

    Set different line location between consecutive debugger stops.

    By default, the debugger traces all events possible including line,
    exceptions, call and return events. Just this alone may mean that for
    any given source line several consecutive stops at a given line may
    occur. Independent of this, Python allows one to put several commands
    in a single source line of code. When a programmer does this, it might
    be because the programmer thinks of the line as one unit.

    One of the challenges of debugging is getting the granualarity of
    stepping comfortable. Because of the above, stepping all events can
    often be too fine-grained and annoying. By setting different on you
    can set a more coarse-level of stepping which often still is small
    enough that you won't miss anything important.
    Note that the `step` and `next` debugger commands have `+` and `-`
    suffixes if you wan to override this setting on a per-command basis.

    See also:
    ---------
    `set trace` to change what events you want to filter.
    `show trace`"""

    in_list = True
    min_abbrev = len("dif")  # Min is "set dif"
    pass


if __name__ == "__main__":
    from trepan.processor.command.set_subcmd import __demo_helper__ as Mhelper

    Mhelper.demo_run(SetDifferent)
    pass



#!/usr/bin/python
"""Greatest Common Divisor

Some characterstics of this program used for testing check_args() does
not have a 'return' statement.

check_args() raises an uncaught exception when given the wrong number
of parameters.

"""
import sys

from trepan.api import debug


def check_args():
    if len(sys.argv) != 3:
        # Rather than use sys.exit let's just raise an error
        raise Exception("Need to give two numbers")
    for i in range(2):
        try:
            sys.argv[i + 1] = int(sys.argv[i + 1])
        except ValueError:
            print("** Expecting an integer, got: %s" % repr(sys.argv[i]))
            sys.exit(2)


def gcd(a, b):
    """GCD. We assume positive numbers"""

    # Make: a <= b
    if a > b:
        (a, b) = (b, a)

    if a <= 0:
        debug(step_ignore=0)
        return None
    if a == 1 or b - a == 0:
        debug()
        return a
    return gcd(b - a, a)


if __name__ == "__main__":
    check_args()

    (a, b) = sys.argv[1:3]
    print("The GCD of %d and %d is %d" % (a, b, gcd(a, b)))



# -*- coding: utf-8 -*-
#   Copyright (C) 2012-2013, 2015-2016, 2020 Rocky Bernstein
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from pyficache import clear_file_format_cache

# Our local modules
from trepan.processor.command.base_subcmd import DebuggerSubcommand
from trepan.lib.complete import complete_token
from trepan.lib.format import color_tf


class SetHighlight(DebuggerSubcommand):
    """**set highlight** [ **reset** ] {**plain** | **light** | **dark** | **off**}

    Set whether we use terminal highlighting. Permissible values are:

           plain:  no terminal highlighting
           off:    same as plain
           light:  terminal background is light (the default)
           dark:   terminal background is dark

    If the first argument is *reset*, we clear any existing color formatting
    and recolor all source code output.

    A related setting is *style* which sets the Pygments style for terminal
    that support, 256 colors. But even here, it is useful to set
    the highlight to tell the debugger for bold and emphasized text what
    values to use.

    Examples:
    --------

        set highlight off   # no highlight
        set highlight plain # same as above
        set highlight       # same as above
        set highlight dark  # terminal has dark background
        set highlight light # terminal has light background
        set highlight reset light # clear source-code cache and
                                  # set for light background
        set highlight reset # clear source-code cache

    See also:
    ---------
    `show highlight` and `set style`"""

    in_list = True
    min_abbrev = len("hi")
    short_help = "Set whether we use terminal highlighting"

    highlight_choices = ("reset", "plain", "light", "dark", "off")

    def complete(self, prefix):
        return complete_token(SetHighlight.highlight_choices, prefix)

    def get_highlight_type(self, arg):
        if not arg:
            return "light"
        if arg in SetHighlight.highlight_choices:
            return arg
        else:
            self.errmsg(
                'Expecting %s"; got %s'
                % (", ".join(SetHighlight.highlight_choices), arg)
            )
            return None
        pass

    def run(self, args):
        if len(args) >= 1 and "reset" == args[0]:
            if len(args) >= 2:
                highlight_type = self.get_highlight_type(args[1])
            else:
                highlight_type = self.debugger.settings[self.name]
            if not highlight_type:
                return
            clear_file_format_cache()
        elif len(args) == 0:
            highlight_type = "plain"
        else:
            highlight_type = self.get_highlight_type(args[0])
            if not highlight_type:
                return
            if "off" == highlight_type:
                highlight_type = "plain"
            else:
                clear_file_format_cache()
            pass
        self.debugger.settings[self.name] = highlight_type
        if highlight_type in ("dark", "light"):
            color_tf.bg = highlight_type
        self.proc.set_prompt()
        show_cmd = self.proc.commands["show"]
        show_cmd.run(["show", self.name])
        return

    pass


# if __name__ == '__main__':
#     from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run
#     demo_run(SetHighlight)
#     pass



"""
Functional tests. In contrast to most unit tesets,
these tests work with a Trepan debugger object and traces or debugs some code with it, testing some particular
aspect of debugging, such as the ability to step, skip or jump to debugged lines, hit breakpoints, etc.
"""



# -*- coding: utf-8 -*-
#
#    Copyright (C) 2009, 2012-2013, 2015, 2020, 2023-2024
#    Rocky Bernstein
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
#    02110-1301 USA.
import glob
import os.path as osp
import re

from trepan.lib.complete import complete_token, complete_token_filtered
from trepan.misc import wrapped_lines
from trepan.processor import cmdproc as Mcmdproc

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand

categories = {
    "breakpoints": "Making the program stop at certain points",
    "data": "Examining data",
    "files": "Specifying and examining files",
    "running": "Running the program",
    "status": "Status inquiries",
    "support": "Support facilities",
    "stack": "Examining the call stack",
    "syntax": "Debugger command syntax",
}


class HelpCommand(DebuggerCommand):
    """**help** [*command* [*subcommand*]|*expression*]

    Without argument, print the list of available debugger commands.

    When an argument is given, it is first checked to see if it is command
    name.

    If the argument is an expression or object name, you get the same
    help that you would get inside a Python shell running the built-in
    *help()* command.

    If the environment variable *$PAGER* is defined, the file is
    piped through that command.  You'll notice this only for long help
    output.

    Some commands like `info`, `set`, and `show` can accept an
    additional subcommand to give help just about that particular
    subcommand. For example `help info line` give help about the
    info line command.

    See also:
    ---------

    `examine` and `whatis`."""

    aliases = ("?",)
    short_help = "Print commands or give help for command(s)"
    HELP_DIR = osp.join(osp.dirname(__file__), "help")
    RST_EXTENSION = ".rst"

    DebuggerCommand.setup(
        locals(),
        category="support",
    )

    def complete(self, prefix):
        proc_obj = self.proc
        matches = complete_token(
            list(categories.keys()) + ["*", "all"] + list(proc_obj.commands.keys()),
            prefix,
        )
        aliases = complete_token_filtered(proc_obj.aliases, prefix, matches)
        return sorted(matches + aliases)

    def run(self, args):
        # It does not make much sense to repeat the last help
        # command. Also, given that 'help' uses PAGER, the you may
        # enter an extra CR which would rerun the (long) help command.
        self.proc.last_command = ""

        if len(args) > 1:
            cmd_name = args[1]
            if cmd_name == "*":
                self.section("List of all debugger commands:")
                m = self.columnize_commands(list(self.proc.commands.keys()))
                self.msg_nocr(m)
                return
            elif cmd_name == "aliases":
                self.show_aliases()
                return
            elif cmd_name == "macros":
                self.show_macros()
                return
            elif cmd_name == "syntax":
                self.show_command_syntax(args)
                return
            elif cmd_name in list(categories.keys()):
                self.show_category(cmd_name, args[2:])
                return

            command_name = Mcmdproc.resolve_name(self.proc, cmd_name)
            if command_name:
                instance = self.proc.commands[command_name]
                if hasattr(instance, "help"):
                    return instance.help(args)
                else:
                    doc = instance.__doc__ or instance.run.__doc__
                    doc = doc.rstrip("\n")
                    self.rst_msg(doc.rstrip("\n"))
                    aliases = [
                        key
                        for key in self.proc.aliases
                        if command_name == self.proc.aliases[key]
                    ]
                    if len(aliases) > 0:
                        self.msg("")
                        msg = wrapped_lines(
                            "Aliases:", ", ".join(aliases) + ".", self.settings["width"]
                        )
                        self.msg(msg)
                        pass
                    pass
            else:
                cmds = [
                    cmd
                    for cmd in list(self.proc.commands.keys())
                    if re.match("^" + cmd_name, cmd)
                ]
                if cmds is None:
                    self.errmsg(
                        f'No commands found matching /^{cmd_name}/. Try "help".'
                    )
                elif len(cmds) == 1:
                    self.msg(f"Pattern '{cmd_name}' matches command {cmds[0]}...")
                    args[1] = cmds[0]
                    self.run(args)
                else:
                    self.section(f"Command names matching /^{cmd_name}/:")
                    self.msg_nocr(self.columnize_commands(cmds))
                    pass
            return
        else:
            self.list_categories()
            pass

        return False

    def list_categories(self):
        """List the command categories and a short description of each."""
        self.section("Classes of commands:")
        cats = list(categories.keys())
        cats.sort()
        for cat in cats:  # Foo! iteritems() doesn't do sorting
            self.msg("  %-13s -- %s" % (cat, categories[cat]))
            pass
        final_msg = """
Type `help` followed by a class name for a list of commands in that class.
Type `help aliases` for a list of current aliases.
Type `help macros` for a list of current macros.
Type `help syntax *item*` for help on syntax *item*
Type `help *` for the list of all commands.
Type `help` *regexp* for the list of commands matching /^#{*regexp*}/
Type `help` followed by command name for full documentation.
"""
        for line in re.compile("\n").split(final_msg.rstrip("\n")):
            self.rst_msg(line)
            pass
        return

    def show_category(self, category, args):
        """Show short help for all commands in `category'."""
        n2cmd = self.proc.commands
        names = list(n2cmd.keys())
        if len(args) == 1 and args[0] == "*":
            self.section(f"Commands in class {category}:")
            cmds = [cmd for cmd in names if category == n2cmd[cmd].category]
            cmds.sort()
            self.msg_nocr(self.columnize_commands(cmds))
            return

        self.msg(f"{categories[category]}.\n")
        self.section("List of commands:")
        names.sort()
        for name in names:  # Foo! iteritems() doesn't do sorting
            if category != n2cmd[name].category:
                continue
            self.msg(
                "%-13s -- %s"
                % (
                    name,
                    n2cmd[name].short_help,
                )
            )
            pass
        return

    def syntax_files(self):
        path = osp.join(self.HELP_DIR, (f"*{self.RST_EXTENSION}"))
        files = glob.glob(path)
        return [osp.basename(name).split(".")[0] for name in files]

    def show_aliases(self):
        self.section("All alias names:")
        m = self.columnize_commands(list(sorted(self.proc.aliases.keys())))
        self.msg_nocr(m)

    def show_macros(self):
        self.section("All macro names:")
        m = self.columnize_commands(list(sorted(self.proc.macros.keys())))
        self.msg_nocr(m)

    def init_syntax_summary_help(self):
        self.syntax_summary_help = {}
        self.syntax_help = {}
        for name in self.syntax_files():
            path = osp.join(self.HELP_DIR, f"{name}{self.RST_EXTENSION}")
            self.syntax_help[name] = "".join(open(path).readlines())
            self.syntax_summary_help[name] = open(path).readline().strip()
            pass
        return

    def show_command_syntax(self, args):
        if not hasattr(self, "syntax_summary_help"):
            self.init_syntax_summary_help()
            pass
        if len(args) == 2 or len(args) == 3 and args[2] == "*":
            self.section("List of syntax help")
            for name, help in self.syntax_summary_help.items():
                self.msg("  %-8s -- %s" % (name, help))
                pass
        else:
            for name in args[2:]:
                if name in self.syntax_files():
                    self.rst_msg(self.syntax_help[name])
                else:
                    self.errmsg(f"No syntax help for {name}")
                pass
            pass
        return

    pass


if __name__ == "__main__":
    # Demo it.
    from trepan import debugger

    d = debugger.Trepan()
    cp = d.core.processor

    command = HelpCommand(cp)
    print("-" * 20)
    command.run(["help"])
    print("-" * 20)
    command.run(["help", "*"])
    print("-" * 20)
    command.run(["help", "quit"])
    # print('-' * 20)
    # command.run(['help', 'stack'])
    # print('-' * 20)
    # command.run(['help', 'breakpoints'])
    print("-" * 20)
    command.run(["help", "breakpoints", "*"])
    print("-" * 20)
    # command.run(['help', 'c.*'])
    # print('-' * 20)
    command.show_command_syntax(["help", "syntax"])
    command.show_command_syntax(["help", "syntax", "command"])
    pass



"""Unit test for trepan.processor.command.cmdfns"""

import pytest

from trepan.processor.cmdfns import get_an_int, get_int, get_onoff, want_different_line

errors = []


def errmsg(msg_str: str):
    errors.append(msg_str)
    return


def reset():
    global errors
    errors = []


def test_get_an_int():
    reset()
    assert 0 == get_an_int(errmsg, "0", "foo", 0)
    assert 0 == len(errors)
    assert 6 == get_an_int(errmsg, "6*1", "foo", 5)
    assert 0 == len(errors)
    assert get_an_int(errmsg, "0", "0 is too small", 5) is None
    assert 1 == len(errors)
    assert get_an_int(errmsg, "4+a", "4+a is invalid", 5) is None
    assert "4+a is invalid" == errors[-1]
    return


def test_get_int():
    reset()
    assert 1 == get_int(errmsg, "1", 5)
    assert 3 == get_int(errmsg, "1+2", 5)
    assert 5 == get_int(errmsg, None, 5)
    assert 1 == get_int(errmsg, None)
    with pytest.raises(ValueError):
        get_int(errmsg, "Foo", 5)
    return


def test_get_onoff():
    for arg in ("1", "on"):
        assert get_onoff(errmsg, arg) is True
        pass
    for arg in ("0", "off"):
        assert get_onoff(errmsg, arg) is False
        pass
    for result in (True, False):
        assert result == get_onoff(errmsg, None, result)
        pass
    with pytest.raises(ValueError):
        get_onoff(errmsg, "Foo")
    return


def test_want_different_line():
    for cmd, default, expected in [
        ("s+", False, True),
        ("s-", True, False),
        ("s", False, False),
        ("n", True, True),
    ]:
        assert expected == want_different_line(cmd, default), cmd
        pass
    return



# -*- coding: utf-8 -*-
#  Copyright (C) 2017-2018, 2020-2021, 2023 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Our local modules
from sys import version_info

from xdis import IS_PYPY, PYTHON_VERSION_TRIPLE

from trepan.processor.command.base_cmd import DebuggerCommand

if PYTHON_VERSION_TRIPLE >= (3, 9):
    pass
else:
    if (3, 7) <= PYTHON_VERSION_TRIPLE:
        from decompyle3.semantics.fragments import deparse_code, deparsed_find
    else:
        from uncompyle6.semantics.fragments import deparse_code, deparsed_find

    class DEvalCommand(DebuggerCommand):
        """**deval**
            **deval?**

        Run a the current deparsed expression in the context of the current
        frame. Normally we are stopped before an expression so the thing that
        corresponds to the `eval` command is running the parent
        construct. `deval?` will run just the command associated with the next
        piece of code to be run.

        Examples:
        ---------

            deval   # Run *parent* of current deparsed code
            deval?  # Run current deparsed code

        See also:
        ---------

        `eval`

        """

        aliases = ("deval?",)
        short_help = "Print value of deparsed expression"

        DebuggerCommand.setup(locals(), category="data", need_stack=True, max_args=0)

        def run(self, args):
            co = self.proc.curframe.f_code
            name = co.co_name
            last_i = self.proc.curframe.f_lasti
            if last_i == -1:
                last_i = 0

            sys_version = version_info[0] + (version_info[1] / 10.0)
            try:
                deparsed = deparse_code(sys_version, co, is_pypy=IS_PYPY)
                nodeInfo = deparsed_find((name, last_i), deparsed, co)
                if not nodeInfo:
                    self.errmsg("Can't find exact offset %d" % last_i)
                    return
            except Exception:
                self.errmsg("error in deparsing code")
                return
            if "?" == args[0][-1]:
                extractInfo = deparsed.extract_node_info(nodeInfo.node)
            else:
                extractInfo, _ = deparsed.extract_parent_info(nodeInfo.node)
            text = extractInfo.selectedText
            text = text.strip()
            self.msg("Evaluating: %s" % text)
            try:
                self.proc.exec_line(text)
            except Exception:
                pass


if __name__ == "__main__":
    import inspect

    from trepan import debugger

    d = debugger.Trepan()
    cp = d.core.processor
    cp.curframe = inspect.currentframe()
    command = DEvalCommand(cp)
    me = 10

    pass



# -*- coding: utf-8 -*-
#   Copyright (C) 2015, 2017-2018, 2020, 2024 Rocky Bernstein
#

from pygments.styles import STYLE_MAP
from trepan.lib.complete import complete_token
from trepan.processor.command.base_subcmd import DebuggerSubcommand

style_names = sorted(list(STYLE_MAP.keys()))


class SetStyle(DebuggerSubcommand):
    """**set style** [*pygments-style*]

    Set the pygments style in to use in formatting text for a 256-color terminal.
    Note: if your terminal doesn't support 256 colors, you may be better off
    using `--highlight=plain` or `--highlight=dark` instead. To turn off styles
    use `set style none`.

    To list the available pygments styles inside the debugger, omit the style name.

    Examples:
    ---------

        set style            # give a list of the style names
        set style colorful   # Pygments 'colorful' style
        set style none       # Turn off style, still use highlight though

    See also:
    --------

    `show style`, `set highlight`"""

    def complete(self, prefix):
        return complete_token(style_names, prefix)

    in_list = True
    min_abbrev = len("sty")
    short_help = "Set the pygments style"

    def run(self, args):
        if len(args) == 0:
            return self.proc.commands["show"].cmds.subcmds["styles"].run([])
        if args[0] == "none":
            self.debugger.settings[self.name] = None
            return

        if args[0] not in style_names:
            self.errmsg(f"style name '{args[0]}' not valid. valid names are: ")
            self.msg(self.columnize_commands(style_names))
            return

        self.debugger.settings[self.name] = args[0]
        return

    pass


if __name__ == '__main__':
    from trepan.processor.command.set_subcmd.__demo_helper__ import demo_run
    demo_run(SetStyle, ["emacs"])
    demo_run(SetStyle, [])
    pass



"""Unit test for trepan.clifns"""
import inspect
import os
import sys

import pytest

from trepan.clifns import is_ok_line_for_breakpoint, path_expanduser_abs


def test_clifns():
    """Test trepan.clifns.expanduser_abs()"""
    file1 = os.path.join(os.path.curdir, "test_clifns")
    file1 = path_expanduser_abs(file1)
    file2 = path_expanduser_abs("test_clifns")
    assert file1 == file2, "path_expanduser_abs"
    assert path_expanduser_abs("~/foo")
    return


def test_is_ok_line_for_breakpoint():
    """Test trepan.clifns.is_ok_for_breakpoint()"""
    filename = __file__
    if len(filename) > 4 and filename[-4:] == ".pyc":
        filename = filename[:-1]
        pass

    # Pick up a Python code line for testing.
    # Note that this comment line relative to the line
    # we pick up is also used.
    frame = inspect.currentframe()
    pytest.mark.skipif(frame is None, reason="Can't get runtime stack frame")
    assert frame is not None
    lineno = frame.f_lineno
    assert is_ok_line_for_breakpoint(filename, lineno, sys.stdout.write)
    assert not is_ok_line_for_breakpoint(filename, lineno - 5, sys.stdout.write)



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2008-2009, 2013, 2015, 2018, 2020, 2023 Rocky
#   Bernstein <rocky@gnu.org>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
import re

from getopt import getopt, GetoptError

# Our local modules
from trepan.processor.command import base_subcmd as Mbase_subcmd
from trepan.lib import pp as Mpp
from trepan.lib import complete as Mcomplete

# when the "with" statement is used, there
# can be get variables having names
# _[1], _[2], etc.
_with_local_varname = re.compile(r"_\[[0-9+]\]")


class InfoLocals(Mbase_subcmd.DebuggerSubcommand):
    """**info locals** [-l | --list | | -h --help]

        **info locals** [*var1 ...*]

    **info locals** *

    With no arguments, show all of the local variables of the current stack
    frame. If a list of names is provide limit display to just those
    variables.

    If `*` is given, just show the variable names, not the values.

    See also:
    ---------
    `info globals`, `info args`, `info frame`"""

    min_abbrev = 2
    need_stack = True
    short_help = "Show the local variables of current stack frame"

    def complete(self, prefix):
        completions = sorted(["*"] + self.proc.curframe.f_locals.keys())
        return Mcomplete.complete_token(completions, prefix)

    def run(self, args):
        if not self.proc.curframe:
            self.errmsg("No frame selected")
            return False

        try:
            opts, args = getopt(
                args,
                "hl",
                ["help", "list"],
            )
        except GetoptError as err:
            # print help information and exit:
            self.errmsg(
                str(err)
            )  # will print something like "option -a not recognized"
            return

        list_only = False
        for o, a in opts:
            if o in ("-h", "--help"):
                self.proc.commands["help"].run(["help", "info", "locals"])
                return
            elif o in ("-l", "--list"):
                list_only = True
            else:
                self.errmsg("unhandled option '%s'" % o)
            pass
        pass

        names = list(self.proc.curframe.f_locals.keys())

        if list_only:
            for name in names:
                self.msg(name)
            return
        if len(args) > 0 and args[0] == "*":
            self.section("locals")
            self.msg(self.columnize_commands(names))
        elif len(args) == 0:
            for name in sorted(names):
                # ALB: a fix for a problem with the new 'with'
                # statement. It seems to work, but I don't know exactly
                # why... (the problem was in self.getval called by
                # info_locals)
                if _with_local_varname.match(name):
                    val = self.proc.curframe.f_locals[name]
                else:
                    val = self.proc.getval(name)
                    pass
                Mpp.pp(
                    val,
                    self.settings["width"],
                    self.msg_nocr,
                    self.msg,
                    prefix="%s =" % name,
                )
                pass
            pass
        else:
            for name in args:
                # ALB: a fix for a problem with the new 'with'
                # statement. It seems to work, but I don't know exactly
                # why... (the problem was in self.getval called by
                # info_locals)
                if name in names:
                    if _with_local_varname.match(name):
                        val = self.proc.curframe.f_locals[name]
                    else:
                        val = self.proc.getval(name)
                        pass
                    Mpp.pp(
                        val,
                        self.settings["width"],
                        self.msg_nocr,
                        self.msg,
                        prefix="%s =" % name,
                    )
                else:
                    self.errmsg("%s is not a local variable" % name)
                    pass
        return False

    pass


if __name__ == "__main__":
    from trepan.processor.command import mock, info as Minfo
    from trepan import debugger as Mdebugger

    d = Mdebugger.Trepan()
    d, cp = mock.dbg_setup(d)
    i = Minfo.InfoCommand(cp)
    sub = InfoLocals(i)
    import inspect

    cp.curframe = inspect.currentframe()
    sub.run([])
    sub.run(["*"])
    sub.run(["Minfo"])
    pass



"""Unit test for trepan.processor.command.disassemble"""
import inspect
from test.unit.cmdhelper import setup_unit_test_debugger

import pytest

from trepan.processor.command.disassemble import DisassembleCommand

msgs = []
errmsgs = []
last_was_newline = True


# FIXME: put in a more common place
# Possibly fix up Mock to include this
def setup_io(command):
    clear_output()
    command.msg = msg
    command.errmsg = errmsg
    command.msg_nocr = msg_nocr
    return


def clear_output():
    global msgs, errmsgs, last_was_newline
    msgs = []
    errmsgs = []
    last_was_newline = True
    return


def msg(msg_str: str):
    msg_nocr(msg_str)
    global last_was_newline
    last_was_newline = True
    return


def msg_nocr(msg_str: str):
    global last_was_newline
    if last_was_newline:
        msgs.append("")
        pass
    msgs[-1] += msg_str
    last_was_newline = len(msg_str) == 0
    return


def errmsg(msg):
    errmsgs.append(msg)
    pass


def test_disassemble():
    """Test processor.command.disassemble.run()"""

    d, cp = setup_unit_test_debugger()
    command = DisassembleCommand(cp)

    setup_io(command)

    pytest.mark.skip("Skipping until disassembly revamp complete")
    return

    command.run(["disassemble"])
    assert len(errmsgs) > 0
    assert len(msgs) == 0
    me = test_disassemble  # NOQA
    cp.curframe = inspect.currentframe()
    # All of these should work
    for args in (
        ["disassemble"],
        ["disassemble", "cp.errmsg"],
        ["disassemble", "unittest"],
        ["disassemble", "1"],
        ["disassemble", "10", "100"],
        ["disassemble", "*10", "*30"],
        ["disassemble", "+", "1"],
        ["disassemble", "-", "1"],
        ["disassemble", "1", "2"],
        ["disassemble", "me"],
    ):
        clear_output()
        try:
            command.run(args)
        except NotImplementedError:
            return

        assert len(msgs) > 0, "msgs for: %s" % " ".join(args)
        assert len(errmsgs) == 0, "errmsgs for: %s %s" % (
            " ".join(args),
            "\n".join(errmsgs),
        )
        pass
    return



"""Unit test for trepan.lib.file"""
import sys

from pygments.lexers import RstLexer

from trepan.lib.format import MonoRSTTerminalFormatter, RstFilter, highlight


def test_mono():
    rst_lex = RstLexer()
    rst_filt = RstFilter()
    rst_lex.add_filter(rst_filt)
    rst_tf = MonoRSTTerminalFormatter()
    text = "`A` very *emphasis* **strong** `code`"
    if sys.version_info[0:2] <= (3, 2):
        return
    got = highlight(text, rst_lex, rst_tf)
    assert '"A" very *emphasis* STRONG "code" ', got

    quit_text = "QUIT - gently terminate the debugged program. "
    rst_tf.reset(80)
    got = highlight(quit_text, rst_lex, rst_tf)
    assert got.startswith(quit_text)

    text = """
This is an example to show off *reformatting.*
We have several lines
here which should be reflowed.

But paragraphs should be respected.

    And verbatim
    text should not be
    touched.

End of test.
"""
    rst_tf.reset(30)
    got = highlight(text, rst_lex, rst_tf)
    assert (
        "This is an example to show \n"
        "off *reformatting.* We have \n"
        "several lines here which \n"
        "should be reflowed. \n\n"
        "But paragraphs should be \n"
        "respected. \n\n"
        "    And verbatim\n"
        "    text should not be\n"
        "    touched.\n\n\n"
        "End of test. " == got
    )
    return



# -*- coding: utf-8 -*-
#  Copyright (C) 2015, 2020 Rocky Bernstein
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import inspect

# Our local modules
from trepan.processor.command.base_cmd import DebuggerCommand
from trepan.processor.complete import complete_bpnumber


class ClearCommand(DebuggerCommand):
    """**clear** [*linenumber*]

    Clear some breakpoints by line number.

    See also:
    ---------
    `delete`
    """

    short_help = "Delete some breakpoints on a line"

    DebuggerCommand.setup(locals(), category="breakpoints", need_stack=True)

    complete = complete_bpnumber

    def run(self, args):
        proc = self.proc
        curframe = proc.curframe
        filename = proc.list_filename
        if len(args) <= 1:
            lineno = inspect.getlineno(curframe)
        else:
            lineno = proc.get_an_int(
                args[1],
                "The 'clear' command argument when given should be "
                "a line number. Got %s",
                min_value=1,
            )
            if lineno is None:
                return

        linenos = self.core.bpmgr.delete_breakpoints_by_lineno(filename, lineno)
        if len(linenos) == 0:
            self.errmsg("No breakpoint at line %d" % lineno)
        elif len(linenos) == 1:
            self.msg("Deleted breakpoint %d" % linenos[0])
        elif len(linenos) > 1:
            self.msg("Deleted breakpoints %s" % " ".join([str(i) for i in linenos]))
        return


if __name__ == "__main__":
    from trepan.debugger import Trepan
    from trepan.processor.cmdproc import get_stack

    d = Trepan()
    cp = d.core.processor
    cp.curframe = inspect.currentframe()
    cp.stack, cp.curindex = get_stack(cp.curframe, None, None, cp)
    command = ClearCommand(d.core.processor)
    command.run(["clear"])
    pass



"""Unit test for trepan.processor.command.break"""

import os
import platform
from test.unit.cmdhelper import errmsg, msg, reset_output, setup_unit_test_debugger

from xdis import PYTHON_VERSION_TRIPLE

from trepan.processor.cmdbreak import parse_break_cmd

# We have to use this subterfuge because "break" is Python reserved word,
# so it can't be used as a module-name component.
# Python "import" irregularities strike again!
break_module = __import__("trepan.processor.command.break", None, None, ["*"])


def parse_break_cmd_wrapper(proc, cmd):
    proc.current_command = cmd
    args = cmd.split(" ")
    return parse_break_cmd(proc, args)


def test_parse_break_cmd():
    reset_output()

    _, cp = setup_unit_test_debugger()
    cmd = break_module.BreakCommand(cp)
    cmd.msg = msg
    cmd.errmsg = errmsg
    proc = cmd.proc

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "break")
    assert fi.endswith("test_cmd_break.py")
    assert (None, None, True, True) == (fn, cond, li > 1, offset > 0)

    assert fi.endswith(__file__)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "break 11")
    assert (None, None, 11, None) == (fn, cond, li, offset)

    if platform.system() == "Windows":
        brk_cmd = f'b """{__file__}""":8'
    else:
        brk_cmd = f"b {__file__}:8"

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, brk_cmd)

    assert (True, 8) == (isinstance(fi, str), li)
    # FIXME: This varies. Why?
    # assert "<module>" == fn

    def foo():
        return "bar"

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "break foo()")
    assert (foo, True, True) == (fn, fi.endswith(__file__), li > 1)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "break food()")
    assert (None, None, None, None) == (fn, fi, li, cond)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "b os.path:5")
    assert (os.path, True, 5) == (fn, isinstance(fi, str), li)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "b os.path.join()")
    assert (os.path.join, True, True) == (fn, isinstance(fi, str), li > 1)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "break if True")
    assert (None, True, True) == (fn, fi.endswith(__file__), li > 1)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "b foo() if True")
    assert (foo, True, True) == (fn, fi.endswith(__file__), li > 1)

    fn, fi, li, cond, offset = parse_break_cmd_wrapper(proc, "br os.path:10 if True")
    assert (True, 10) == (isinstance(fi, str), li)

    # FIXME:
    # Try:
    #  a breakpoint with a symlink in the filename.
    #  breakpoint with a single quotes and embedded black
    #  breakpoint with a double quotes and embedded \,
    #  Triple quote things
    #
    # Also, add a unit test for canonic.



"""Unit test for trepan.lib.pp"""

from typing import List

from trepan.lib.pp import pp, pprint_simple_array

errmsgs: List[str] = []


def errmsg(msg_str: str):
    """
    error message collection routine for unit testing.
    """
    global errmsgs
    errmsgs.append(msg_str)
    return


msgs: List[str] = []


def msg(msg_str: str):
    """
    message collection routine for unit testing.
    """
    global msgs
    msgs.append(msg_str)
    return


def reset_output():
    """
    Reset error messages and messages in advance of testing.
    """
    global errmsgs
    errmsgs = []
    global msgs
    msgs = []


def msg_nocr(msg_str: str):
    global msgs
    if len(msgs) > 0:
        msgs[-1] += msg_str
    else:
        msgs = [msg_str]
    return


def test_lib_pp():
    reset_output()
    pp([i for i in range(8)], 50, msg_nocr, msg)
    assert ["[0, 1, 2, 3, 4, 5, 6, 7]\n\n"] == msgs

    reset_output()
    x = [i for i in range(5)]
    pp(x, 50, msg_nocr, msg, "x =")
    assert ["x = [0, 1, 2, 3, 4]"] == msgs

    reset_output()
    x = [i for i in range(10)]
    pp(x, 20, msg_nocr, msg, "x = ")
    assert ["x = [0, 1, 2, 3, 4, 5,\n 6, 7, 8, 9]\n\n"] == msgs
    return


def test_lib_pprint_simple_array():
    reset_output()

    # We can't pprint these types, so return False.
    for val in (5, {}, None, True):
        assert pprint_simple_array(5, 53, msg_nocr, msg) is False

    pprint_simple_array(list(range(50)), 53, msg_nocr, msg)
    assert [
        "[ 0,  1,  2,  3,  4,  5,  6,  7,  8,  9, 10, 11,\n"
        " 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23,\n"
        " 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,\n"
        " 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47,\n"
        " 48, 49]\n\n"
    ] == msgs
    return



"""Unit test for trepan.lib.thread"""
import _thread
import sys
import threading

import pytest

from trepan.lib.thred import (
    current_thread_name,
    find_debugged_frame,
    id2thread_name,
    map_thread_names,
)


class BgThread(threading.Thread):
    def __init__(self, name_checker):
        threading.Thread.__init__(self)
        self.id_name_checker = name_checker
        return

    def run(self):
        self.id_name_checker()
        return


def id_name_checker():
    """Helper for testing map_thread_names and id2thread"""
    name2id = map_thread_names()
    for thread_id, f in list(sys._current_frames().items()):
        assert thread_id == name2id[id2thread_name(thread_id)]
        # FIXME: use a better test
        assert f != find_debugged_frame(f)
        pass


def test_current_thread_name():
    """Test trepan.lib.thred.current_thread_name"""
    assert "MainThread" == current_thread_name()


@pytest.mark.skip(reason="Fix have an intermittent failure")
def test_id2thread_name():
    """Test ``map_thread_names`` and ``id2thread``."""
    thread_id = _thread.get_ident()
    assert "MainThread" == id2thread_name(thread_id)
    id_name_checker()

    background = BgThread(id_name_checker)
    background.start()
    background.join()  # Wait for the background task to finish



"""Unit test for trepan.processor.command.alias and unalias"""
import inspect
from test.unit.cmdhelper import setup_unit_test_debugger


def test_alias_unalias_command():
    """Test 'alias' and 'unalias' commands"""
    errors = []
    msgs = []

    def errmsg(msg_str: str):
        errors.append(msg_str)
        return

    def msg(msg_str: str):
        msgs.append(msg_str)
        return

    def check_alias(should_not_have, cmd_name, *args):
        cmdproc.msgs = []
        cmdproc.errmsgs = []
        cmds = cmdproc.commands
        arg_str = " ".join(args)
        my_cmd = cmds[cmd_name]
        newargs = [cmd_name]
        newargs += args
        my_cmd.run(newargs)

        if should_not_have:
            shoulda = ["", "no "]
        else:
            shoulda = ["no ", ""]
            pass
        assert should_not_have == (len(msgs) == 0), (
            f"Expecting {shoulda[0]}{cmd_name} for {arg_str}.\n Got {msgs}",
        )

        assert should_not_have != (
            len(errors) == 0
        ), f"Expecting {shoulda[1]}error for #{arg_str}.\n Got {errors}"
        return

    def is_alias_defined(alias_name):
        return alias_name in list(cmdproc.aliases.keys())

    # The usual setup.
    _, cmdproc = setup_unit_test_debugger()
    cmdproc.curframe = inspect.currentframe()
    cmd = cmdproc.commands["alias"]
    cmd.msg = msg
    cmd.errmsg = errmsg
    cmd = cmdproc.commands["unalias"]
    cmd.msg = msg
    cmd.errmsg = errmsg

    assert not len(cmdproc.aliases) == 0, "There should be some aliases defined"

    assert not is_alias_defined("ki")
    check_alias(False, "alias", "ki", "kill")
    assert is_alias_defined("ki")
    check_alias(False, "unalias", "ki")
    assert not is_alias_defined("ki")
    return
